# -*- coding: utf-8 -*-

#-------------------------------------------------------------------------------

# This file is part of Code_Saturne, a general-purpose CFD tool.
#
# Copyright (C) 1998-2021 EDF S.A.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.

#-------------------------------------------------------------------------------

import sys, unittest
from code_saturne.model.XMLvariables import Model
from code_saturne.model.XMLengine import *
from code_saturne.model.XMLmodel import *
from code_saturne.model.MainFieldsModel import MainFieldsModel

from code_saturne.model.ProfilesModel import ProfilesModel
from code_saturne.model.TimeAveragesModel import TimeAveragesModel

#-------------------------------------------------------------------------------
# Constructor
#-------------------------------------------------------------------------------


class NonCondensableModel(MainFieldsModel, Variables, Model):
    """
    This class manages the Field objects in the XML file
    """

    def __init__(self, case):
        """
        Constuctor.
        """
        #
        # XML file parameters
        MainFieldsModel.__init__(self, case)
        self.case            = case
        self.XMLNodethermo   = self.case.xmlGetNode('thermophysical_models')
        self.XMLNodeNonCondensable = self.XMLNodethermo.xmlInitNode('non_condensable_list')
        self.node_anal       = self.case.xmlInitNode('analysis_control')
        self.node_average    = self.node_anal.xmlInitNode('time_averages')
        self.node_profile    = self.node_anal.xmlInitNode('profiles')


    def defaultValues(self):
        default = {}

        default['typeNonCond'] = "Air"
        default['MolarMass']   = 0.02896
        default['Cobin1']      = 2.2e-5
        default['Cobin2']      = 1.5
        return default


    def HydrogenValues(self):
        """
        Return in a dictionnary which contains default values
        """
        hydrogen = {}
        hydrogen['MolarMass'] = 0.002016
        hydrogen['Cobin1'] = 0.78e-4
        hydrogen['Cobin2'] = 1.75
        return hydrogen


    def AirValues(self):
        """
        Return in a dictionnary which contains default values
        """
        air = {}
        air['MolarMass']   = 0.02896
        air['Cobin1']      = 2.2e-5
        air['Cobin2']      = 1.5
        return air


    def NitrogenValues(self):
        """
        Return in a dictionnary which contains default values
        """
        nitrogen = {}
        nitrogen['MolarMass'] = 28.0134e-3
        nitrogen['Cobin1'] = 0.227e-4
        nitrogen['Cobin2'] = 1.75
        return nitrogen


    def HeliumValues(self):
        """
        Return in a dictionnary which contains default values
        """
        helium = {}
        helium['MolarMass'] = 4.003e-3
        helium['Cobin1'] = 0.73e-4
        helium['Cobin2'] = 1.75
        return helium

    def OxygenValues(self):
        """
        Return in a dictionnary which contains default values
        """
        oxygen = {}
        oxygen['MolarMass'] = 32.e-3
        oxygen['Cobin1'] = 0.24e-4
        oxygen['Cobin2'] = 1.71
        return oxygen

    def checkNonCondensableRequirements(self):
        """
        Check if the requirements for non condensable gas addition are met
        - Energy equation must be activated for all gases
        """
        if self.getGasPhaseList() in [[], None]:
            return False
        for field_id in self.getGasPhaseList():
            if self.getEnergyModel(field_id) != "off":
                return True
        return False

    def getNonCondensableLabelList(self):
        """
        Return the non condensable list
        """
        list = []
        for node in self.XMLNodeNonCondensable.xmlGetNodeList('variable'):
            list.append(node['label'])
        return list


    def getNonCondensableNameList(self):
        """
        Return the non condensable list
        """
        list = []
        for node in self.XMLNodeNonCondensable.xmlGetNodeList('variable'):
            list.append(node['name'])
        return list


    def getNonCondensableByFieldId(self, FieldId):
        """
        Return the non condensable name list for a fieldId
        """
        self.isInList(str(FieldId),self.getFieldIdList())
        list = []
        for node in self.XMLNodeNonCondensable.xmlGetNodeList('variable'):
            if self.getNonCondFieldId(node['name']) == str(FieldId) :
                list.append(node['name'])
        return list


    @Variables.undoLocal
    def addNonCondensable(self):
        """
        Add a non condensable
        """
        fieldId = self.getFirstGasField()
        field_name = self.getFieldLabelsList()[int(fieldId)-1]
        type = self.defaultValues()['typeNonCond']
        label   = type + "_1" + "_" + field_name
        if label in self.getNonCondensableLabelList() :
           labelNumber = 1
           label = type + "_" + str(labelNumber) + "_" + field_name
           while label in self.getNonCondensableLabelList() :
               labelNumber += 1
               label = type + "_" + str(labelNumber)+ "_" + field_name

        name = "mass_fraction_non_condensable_gas_" + str(len(self.getNonCondensableNameList())+1)

        Variables(self.case).setNewVariableProperty("variable", "", self.XMLNodeNonCondensable, fieldId, name, label)

        # for non condensable we need use cathare
        from code_saturne.model.ThermodynamicsModel import ThermodynamicsModel
        ThermodynamicsModel(self.case).setMaterials(1, "Water")
        ThermodynamicsModel(self.case).setMethod(1, "Cathare")
        ThermodynamicsModel(self.case).setMaterials(2, "Water")
        ThermodynamicsModel(self.case).setMethod(2, "Cathare")
        del ThermodynamicsModel

        return name


    @Variables.undoLocal
    def setNonCondLabel(self, name, label):
        """
        set label for non condensable
        """
        self.isInList(name,self.getNonCondensableNameList())
        for node in self.XMLNodeNonCondensable.xmlGetNodeList('variable'):
            if node['name'] == name :
               node['label'] = label


    @Variables.noUndo
    def getNonCondLabel(self, name):
        """
        return label for non condensable
        """
        self.isInList(name,self.getNonCondensableNameList())
        for node in self.XMLNodeNonCondensable.xmlGetNodeList('variable'):
            if node['name'] == name :
               return node['label']


    @Variables.undoLocal
    def setNonCondFieldId(self, name, field):
        """
        set field Id for non condensable
        """
        fieldId = -1
        for id in self.getGasPhaseList() :
            if field == self.getLabel(id) :
               fieldId = id
        self.isInList(fieldId, self.getGasPhaseList())
        for node in self.XMLNodeNonCondensable.xmlGetNodeList('variable'):
            if node['name'] == name :
               node['field_id'] = fieldId


    @Variables.noUndo
    def getNonCondFieldId(self, name):
        """
        return field Id for non condensable
        """
        self.isInList(name,self.getNonCondensableNameList())
        for node in self.XMLNodeNonCondensable.xmlGetNodeList('variable'):
            if node['name'] == name :
               return node['field_id']


    @Variables.undoLocal
    def setNonCondType(self, name, type):
        """
        set type for non condensable
        """
        self.isInList(name,self.getNonCondensableNameList())

        for node in self.XMLNodeNonCondensable.xmlGetNodeList('variable'):
            if node['name'] == name :
               childNode = node.xmlInitChildNode('Type')
               childNode.xmlSetAttribute(choice = type)
               self.setIncTyp(name,type)


    @Variables.noUndo
    def getNonCondType(self, name):
        """
        return type for non condensable
        """
        self.isInList(name,self.getNonCondensableNameList())

        for node in self.XMLNodeNonCondensable.xmlGetNodeList('variable'):
            if node['name'] == name :
               childNode = node.xmlGetNode('Type')
               if childNode == None :
                   type = self.defaultValues()['typeNonCond']
                   self.setNonCondType(name, type)
               type = node.xmlGetNode('Type')['choice']
               return type


    @Variables.undoGlobal
    def setIncTyp(self, name, Inctyp):
        masmol = 0.
        cobin1 = 0.
        cobin2 = 0.
        if Inctyp == 'H2' :
            masmol = self.HydrogenValues()['MolarMass']
            cobin1 = self.HydrogenValues()['Cobin1']
            cobin2 = self.HydrogenValues()['Cobin2']
        elif Inctyp == 'N2' :
            masmol = self.NitrogenValues()['MolarMass']
            cobin1 = self.NitrogenValues()['Cobin1']
            cobin2 = self.NitrogenValues()['Cobin2']
        elif Inctyp == 'HE' :
            masmol = self.HeliumValues()['MolarMass']
            cobin1 = self.HeliumValues()['Cobin1']
            cobin2 = self.HeliumValues()['Cobin2']
        elif Inctyp == 'O2' :
            masmol = self.OxygenValues()['MolarMass']
            cobin1 = self.OxygenValues()['Cobin1']
            cobin2 = self.OxygenValues()['Cobin2']
        elif Inctyp == 'Air' :
            masmol = self.AirValues()['MolarMass']
            cobin1 = self.AirValues()['Cobin1']
            cobin2 = self.AirValues()['Cobin2']

        self.setNonCondMassMol(name, masmol)
        self.setNonCondCobin1( name, cobin1)
        self.setNonCondCobin2( name, cobin2)


    @Variables.undoLocal
    def setNonCondMassMol(self, name, MassMol):
        """
        set molar mass for non condensable
        """
        self.isInList(name,self.getNonCondensableNameList())
        self.isPositiveFloat(MassMol)

        for node in self.XMLNodeNonCondensable.xmlGetNodeList('variable'):
            if node['name'] == name :
               node.xmlSetData('MolarMass', MassMol)


    @Variables.noUndo
    def getNonCondMassMol(self, name):
        """
        return mass molar for non condensable
        """
        self.isInList(name,self.getNonCondensableNameList())
        for node in self.XMLNodeNonCondensable.xmlGetNodeList('variable'):
            if node['name'] == name :
               massmol = node.xmlGetDouble('MolarMass')
               if massmol == None :
                   massmol = self.defaultValues()['MolarMass']
                   self.setNonCondMassMol(name, massmol)
               return massmol


    @Variables.undoLocal
    def setNonCondCobin1(self, name, cobin1):
        """
        set cobin 1 for non condensable
        """
        self.isInList(name,self.getNonCondensableNameList())
        self.isPositiveFloat(cobin1)

        for node in self.XMLNodeNonCondensable.xmlGetNodeList('variable'):
            if node['name'] == name :
               node.xmlSetData('Cobin1', cobin1)


    @Variables.noUndo
    def getNonCondCobin1(self, name):
        """
        return cobin 1 for non condensable
        """
        self.isInList(name,self.getNonCondensableNameList())
        for node in self.XMLNodeNonCondensable.xmlGetNodeList('variable'):
            if node['name'] == name :
               cobin1 = node.xmlGetDouble('Cobin1')
               if cobin1 == None :
                   cobin1 = self.defaultValues()['Cobin1']
                   self.setNonCondCobin1(name, cobin1)
               return cobin1


    @Variables.undoLocal
    def setNonCondCobin2(self, name, cobin2):
        """
        set cobin 2 for non condensable
        """
        self.isInList(name,self.getNonCondensableNameList())
        self.isPositiveFloat(cobin2)

        for node in self.XMLNodeNonCondensable.xmlGetNodeList('variable'):
            if node['name'] == name :
               node.xmlSetData('Cobin2', cobin2)


    @Variables.noUndo
    def getNonCondCobin2(self, name):
        """
        return cobin 2 for non condensable
        """
        self.isInList(name,self.getNonCondensableNameList())
        for node in self.XMLNodeNonCondensable.xmlGetNodeList('variable'):
            if node['name'] == name :
               cobin2 = node.xmlGetDouble('Cobin2')
               if cobin2 == None :
                   cobin2 = self.defaultValues()['Cobin2']
                   self.setNonCondCobin2(name, cobin2)
               return cobin2


    @Variables.undoGlobal
    def deleteNonCondensable(self, number):
        """
        Suppress non condensable from list
        """
        name = "mass_fraction_non_condensable_gas_" + str(number)
        self.isInList(name,self.getNonCondensableNameList())

        # delete non condensable
        for node in self.XMLNodeNonCondensable.xmlGetNodeList('variable'):
            try :
               if node['name'] == name :
                  node.xmlRemoveNode()
            except :
               pass

        #suppress profile
        for node in reversed(self.node_profile.xmlGetNodeList('profile')) :
            suppress = 0
            for child in node.xmlGetNodeList('var_prop') :
                if (child['name'] == name) :
                    suppress = 1
            if suppress == 1 :
                name = node['name']
                ProfilesModel(self.case).deleteProfile(name)

        #suppress average
        for node in reversed(self.node_average.xmlGetNodeList('time_average')) :
            suppress = 0
            for child in node.xmlGetNodeList('var_prop') :
                if (child['name'] == name) :
                    suppress = 1
                    break
            if suppress == 1 :
                name = node['name']
                TimeAveragesModel(self.case).deleteTimeAverage(name)

        # update name for other non condensable in XML file
        index = 1
        for node in self.XMLNodeNonCondensable.xmlGetNodeList('variable'):
            try :
               if index >= number :
                  oldname = "mass_fraction_non_condensable_gas_" + str(index+1)
                  name = "mass_fraction_non_condensable_gas_" + str(index)
                  node['name'] = name
                  for n in self.case.xmlGetNodeList('var_prop', name=oldname):
                      n['name'] = name
            except :
               pass
            index += 1

#-------------------------------------------------------------------------------
# DefineUsersScalars test case
#-------------------------------------------------------------------------------

class NonCondensableTestCase(ModelTest):
    """
    """
    def checkNonCondensableInstantiation(self):
        """Check whether the NonCondensableModel class could be instantiated"""
        model = None
        model = NonCondensableModel(self.case)
        assert model != None, 'Could not instantiate NonCondensableModel'


    def checkGetNonCondensableLabelList(self):
        """Check whether the  NonCondensableModel class could get the NonCondensableLabelList"""
        MainFieldsModel(self.case).addDefinedField('1', 'field1', 'continuous', 'gas', 'on', 'on', 'off', 1)
        mdl = NonCondensableModel(self.case)
        mdl.addNonCondensable()
        assert mdl.getNonCondensableLabelList() == ['Air_1'],\
            'Could not get NonCondensableLabelList'


    def checkGetNonCondensableNameList(self):
        """Check whether the  NonCondensableModel class could get the NonCondensableNameList"""
        MainFieldsModel(self.case).addDefinedField('1', 'field1', 'continuous', 'gas', 'on', 'on', 'off', 1)
        mdl = NonCondensableModel(self.case)
        mdl.addNonCondensable()
        assert mdl.getNonCondensableNameList() == ['mass_fraction_non_condensable_gas_1'],\
            'Could not get NonCondensableNameList'


    def checkGetNonCondensableByFieldId(self):
        """Check whether the  NonCondensableModel class could get the NonCondensableByFieldId"""
        MainFieldsModel(self.case).addDefinedField('1', 'field1', 'continuous', 'gas', 'on', 'on', 'off', 1)
        mdl = NonCondensableModel(self.case)
        mdl.addNonCondensable()
        assert mdl.getNonCondensableByFieldId('1') == ['mass_fraction_non_condensable_gas_1'],\
            'Could not get NonCondensableByFieldId'


    def checkaddNonCondensable(self):
        """Check whether the  NonCondensableModel class could addNonCondensable"""
        MainFieldsModel(self.case).addDefinedField('1', 'field1', 'continuous', 'gas', 'on', 'on', 'off', 1)
        mdl = NonCondensableModel(self.case)
        mdl.addNonCondensable()
        doc = '''<non_condensable_list>
                         <variable field_id="1" label="Air_1" name="mass_fraction_non_condensable_gas_1">
                                 <listing_printing status="on"/>
                                 <postprocessing_recording status="on"/>
                         </variable>
                 </non_condensable_list>'''
        assert mdl.XMLNodeNonCondensable == self.xmlNodeFromString(doc),\
            'Could not addNonCondensable'


    def checkGetandSetNonCondLabel(self):
        """Check whether the NonCondensableModel class could set and get NonCondLabel"""
        MainFieldsModel(self.case).addDefinedField('1', 'field1', 'continuous', 'gas', 'on', 'on', 'off', 1)
        mdl = NonCondensableModel(self.case)
        mdl.addNonCondensable()
        mdl.setNonCondLabel('mass_fraction_non_condensable_gas_1','example_label')
        doc = '''<non_condensable_list>
                         <variable field_id="1" label="example_label" name="mass_fraction_non_condensable_gas_1">
                                 <listing_printing status="on"/>
                                 <postprocessing_recording status="on"/>
                         </variable>
                 </non_condensable_list>'''
        assert mdl.XMLNodeNonCondensable == self.xmlNodeFromString(doc),\
            'Could not set NonCondLabel'
        assert mdl.getNonCondLabel('mass_fraction_non_condensable_gas_1') == 'example_label',\
            'Could not get NonCondLabel'


    def checkGetandSetNonCondFieldId(self):
        """Check whether the NonCondensableModel class could set and get NonCondFieldId"""
        MainFieldsModel(self.case).addDefinedField('1', 'field1', 'continuous', 'gas', 'on', 'on', 'off', 1)
        mdl = NonCondensableModel(self.case)
        mdl.addNonCondensable()
        mdl.setNonCondFieldId('mass_fraction_non_condensable_gas_1','field1')
        doc = '''<non_condensable_list>
                         <variable field_id="1" label="Air_1" name="mass_fraction_non_condensable_gas_1">
                                 <listing_printing status="on"/>
                                 <postprocessing_recording status="on"/>
                         </variable>
                 </non_condensable_list>'''
        assert mdl.XMLNodeNonCondensable == self.xmlNodeFromString(doc),\
            'Could not set NonCondFieldId'
        assert mdl.getNonCondFieldId('mass_fraction_non_condensable_gas_1') == '1',\
            'Could not get NonCondFieldId'


    def checkGetandSetNonCondType(self):
        """Check whether the NonCondensableModel class could set and get NonCondType"""
        MainFieldsModel(self.case).addDefinedField('1', 'field1', 'continuous', 'gas', 'on', 'on', 'off', 1)
        mdl = NonCondensableModel(self.case)
        mdl.addNonCondensable()
        mdl.setNonCondType('mass_fraction_non_condensable_gas_1','N2')
        doc = '''<non_condensable_list>
                         <variable field_id="1" label="Air_1" name="mass_fraction_non_condensable_gas_1">
                                 <listing_printing status="on"/>
                                 <postprocessing_recording status="on"/>
                                 <Type choice="N2"/>
                                 <MolarMass>
                                         0.028
                                 </MolarMass>
                                 <Cobin1>
                                         2.27e-05
                                 </Cobin1>
                                 <Cobin2>
                                         1.75
                                 </Cobin2>
                         </variable>
                 </non_condensable_list>'''
        assert mdl.XMLNodeNonCondensable == self.xmlNodeFromString(doc),\
            'Could not set NonCondType'
        assert mdl.getNonCondType('mass_fraction_non_condensable_gas_1') == 'N2',\
            'Could not get NonCondType'


    def checkGetandSetNonCondMassMol(self):
        """Check whether the NonCondensableModel class could set and get NonCondMassMoll"""
        MainFieldsModel(self.case).addDefinedField('1', 'field1', 'continuous', 'gas', 'on', 'on', 'off', 1)
        mdl = NonCondensableModel(self.case)
        mdl.addNonCondensable()
        mdl.setNonCondMassMol('mass_fraction_non_condensable_gas_1',8510.1)
        doc = '''<non_condensable_list>
                         <variable field_id="1" label="Air_1" name="mass_fraction_non_condensable_gas_1">
                                 <listing_printing status="on"/>
                                 <postprocessing_recording status="on"/>
                                 <MolarMass>
                                         8510.1
                                 </MolarMass>
                         </variable>
                 </non_condensable_list>'''
        assert mdl.XMLNodeNonCondensable == self.xmlNodeFromString(doc),\
            'Could not set NonCondMassMol'
        assert mdl.getNonCondMassMol('mass_fraction_non_condensable_gas_1') == 8510.1,\
            'Could not get NonCondMassMol'


    def checkGetandSetNonCondCobin1and2(self):
        """Check whether the NonCondensableModel class could set and get NonCondCobin1and2"""
        MainFieldsModel(self.case).addDefinedField('1', 'field1', 'continuous', 'gas', 'on', 'on', 'off', 1)
        mdl = NonCondensableModel(self.case)
        mdl.addNonCondensable()
        mdl.setNonCondCobin1('mass_fraction_non_condensable_gas_1',5.5)
        mdl.setNonCondCobin2('mass_fraction_non_condensable_gas_1',6.6)
        doc = '''<non_condensable_list>
                         <variable field_id="1" label="Air_1" name="mass_fraction_non_condensable_gas_1">
                                 <listing_printing status="on"/>
                                 <postprocessing_recording status="on"/>
                                 <Cobin1>
                                         5.5
                                 </Cobin1>
                                 <Cobin2>
                                         6.6
                                 </Cobin2>
                         </variable>
                 </non_condensable_list>'''
        assert mdl.XMLNodeNonCondensable == self.xmlNodeFromString(doc),\
            'Could not set NonCondCobin1and2'
        assert mdl.getNonCondCobin1('mass_fraction_non_condensable_gas_1') == (5.5),\
            'Could not get NonCondCobin1'
        assert mdl.getNonCondCobin2('mass_fraction_non_condensable_gas_1') == (6.6),\
            'Could not get NonCondCobin2'


    def checkdeleteNonCondensable(self):
        """Check whether the NonCondensableModel class could deleteNonCondensable"""
        MainFieldsModel(self.case).addDefinedField('1', 'field1', 'continuous', 'gas', 'on', 'on', 'off', 1)
        mdl = NonCondensableModel(self.case)
        mdl.addNonCondensable()
        mdl.addNonCondensable()
        mdl.deleteNonCondensable(1)
        doc = '''<non_condensable_list>
                         <variable field_id="1" label="Air_2" name="mass_fraction_non_condensable_gas_1">
                                 <listing_printing status="on"/>
                                 <postprocessing_recording status="on"/>
                         </variable>
                 </non_condensable_list>'''
        assert mdl.XMLNodeNonCondensable == self.xmlNodeFromString(doc),\
            'Could not delete noncondensable'


    def checksetIncTyp(self):
        """Check whether the NonCondensableModel class could deleteNonCondensable"""
        MainFieldsModel(self.case).addDefinedField('1', 'field1', 'continuous', 'gas', 'on', 'on', 'off', 1)
        mdl = NonCondensableModel(self.case)
        mdl.addNonCondensable()
        mdl.setIncTyp('mass_fraction_non_condensable_gas_1','N2')
        doc = '''<non_condensable_list>
                         <variable field_id="1" label="Air_1" name="mass_fraction_non_condensable_gas_1">
                                 <listing_printing status="on"/>
                                 <postprocessing_recording status="on"/>
                                 <MolarMass>
                                         0.028
                                 </MolarMass>
                                 <Cobin1>
                                         2.27e-05
                                 </Cobin1>
                                 <Cobin2>
                                         1.75
                                 </Cobin2>
                         </variable>
                 </non_condensable_list>'''
        assert mdl.XMLNodeNonCondensable == self.xmlNodeFromString(doc),\
            'Could not setIncTyp for N2'

        mdl.setIncTyp('mass_fraction_non_condensable_gas_1','H2')
        doc = '''<non_condensable_list>
                         <variable field_id="1" label="Air_1" name="mass_fraction_non_condensable_gas_1">
                                 <listing_printing status="on"/>
                                 <postprocessing_recording status="on"/>
                                 <MolarMass>
                                         0.002016
                                 </MolarMass>
                                 <Cobin1>
                                         7.8e-05
                                 </Cobin1>
                                 <Cobin2>
                                         1.75
                                 </Cobin2>
                         </variable>
                 </non_condensable_list>'''
        assert mdl.XMLNodeNonCondensable == self.xmlNodeFromString(doc),\
            'Could not setIncTyp for H2'

        mdl.setIncTyp('mass_fraction_non_condensable_gas_1','HE')
        doc = '''<non_condensable_list>
                         <variable field_id="1" label="Air_1" name="mass_fraction_non_condensable_gas_1">
                                 <listing_printing status="on"/>
                                 <postprocessing_recording status="on"/>
                                 <MolarMass>
                                         0.004003
                                 </MolarMass>
                                 <Cobin1>
                                         7.3e-05
                                 </Cobin1>
                                 <Cobin2>
                                         1.75
                                 </Cobin2>
                         </variable>
                 </non_condensable_list>'''
        assert mdl.XMLNodeNonCondensable == self.xmlNodeFromString(doc),\
            'Could not setIncTyp for HE'

        mdl.setIncTyp('mass_fraction_non_condensable_gas_1','O2')
        doc = '''<non_condensable_list>
                         <variable field_id="1" label="Air_1" name="mass_fraction_non_condensable_gas_1">
                                 <listing_printing status="on"/>
                                 <postprocessing_recording status="on"/>
                                 <MolarMass>
                                         0.032
                                 </MolarMass>
                                 <Cobin1>
                                         2.4e-05
                                 </Cobin1>
                                 <Cobin2>
                                         1.71
                                 </Cobin2>
                         </variable>
                 </non_condensable_list>'''
        assert mdl.XMLNodeNonCondensable == self.xmlNodeFromString(doc),\
            'Could not setIncTyp for O2'

        mdl.setIncTyp('mass_fraction_non_condensable_gas_1','Air')
        doc = '''<non_condensable_list>
                         <variable field_id="1" label="Air_1" name="mass_fraction_non_condensable_gas_1">
                                 <listing_printing status="on"/>
                                 <postprocessing_recording status="on"/>
                                 <MolarMass>
                                         0.02896
                                 </MolarMass>
                                 <Cobin1>
                                         2.2e-05
                                 </Cobin1>
                                 <Cobin2>
                                         1.5
                                 </Cobin2>
                         </variable>
                 </non_condensable_list>'''
        assert mdl.XMLNodeNonCondensable == self.xmlNodeFromString(doc),\
            'Could not setIncTyp for Air'


def suite():
    testSuite = unittest.makeSuite(NonCondensableTestCase, "check")
    return testSuite


def runTest():
    print("NonCondensableTestCase")
    runner = unittest.TextTestRunner()
    runner.run(suite())



#!/usr/bin/env python3

#-------------------------------------------------------------------------------

# This file is part of Code_Saturne, a general-purpose CFD tool.
#
# Copyright (C) 1998-2021 EDF S.A.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.

#-------------------------------------------------------------------------------

"""
This module describes the script used to run a study/case for Code_Saturne.

This module defines the following functions:
- process_cmd_line
- main
"""

#===============================================================================
# Import required Python modules
#===============================================================================

import os, sys, subprocess
import types, string
from argparse import ArgumentParser

from code_saturne import cs_batch
from code_saturne.cs_exec_environment import enquote_arg, get_shell_type
from code_saturne import cs_run
from code_saturne import cs_runcase

#-------------------------------------------------------------------------------
# Print a help page.
#-------------------------------------------------------------------------------

def print_help(submit_cmd, pkg):
    """
    Print a help page.
    """

    help_string = \
"""
Usage: %s [batch options]

Options:
   Any option provided by the batch submission command (%s)
"""
    print(help_string % (sys.argv[0], submit_cmd, pkg.name))

#-------------------------------------------------------------------------------
# Process the command line arguments
#-------------------------------------------------------------------------------

def process_cmd_line(argv, submit_cmd, pkg):
    """
    Process the passed command line arguments.
    """

    runcase_path = None
    need_help = False

    if "-h" in argv or "--help" in argv:
        need_help = True

    if need_help:
        print_help(submit_cmd, pkg)

    return

#===============================================================================
# Sumbit the calculation
#===============================================================================

def main(argv, pkg):
    """
    Main function.
    """

    # Use alternate compute (back-end) package if defined

    batch = cs_batch.batch(pkg)

    submit_cmd = batch.submit_command_prefix()

    if not submit_cmd:
        submit_cmd = get_shell_type()

    epilog = ("Options not listed above are passed to the batch "
              "submission commands; see your batch documentation for this.")

    run_parser = cs_run.arg_parser(argv)
    prog = os.path.basename(sys.argv[0]) + " " + sys.argv[1]

    parser = ArgumentParser(parents=[run_parser],
                            prog=prog,
                            description="Submit a case or specified run stages.",
                            usage='%(prog)s [run options] [batch options]',
                            epilog=epilog,
                            conflict_handler='resolve')

    run_args, submit_args = parser.parse_known_args(argv)

    retcode, result_path, r_c = cs_run.run(pkg=pkg,
                                           run_args=run_args,
                                           submit_args=submit_args)

    # Now prepare runcase for submission

    if retcode != 0 or not result_path:
        err_str = ' run not staged due to prior error.'
        raise Exception(err_str)

    runcase_path = os.path.join(result_path, 'runcase')
    runcase = cs_runcase.runcase(runcase_path,
                                 package=pkg,
                                 submit=True,
                                 job_header = r_c['job_header'],
                                 prologue = r_c['run_prologue'],
                                 epilogue = r_c['run_epilogue'])
    runcase.save()

    # Now submit case

    save_wd = os.getcwd()
    os.chdir(result_path)

    if r_c['job_parameters']:
        submit_cmd += ' ' + r_c['job_parameters']

    submit_cmd += ' ./runcase'

    return subprocess.call(submit_cmd, shell=True)

    os.chdir(save_wd)

#-------------------------------------------------------------------------------

if __name__ == '__main__':

    # Run package
    from code_saturne.cs_package import package
    pkg = package()

    retval = main(sys.argv[1:], pkg)

    sys.exit(retval)

#-------------------------------------------------------------------------------
# End
#-------------------------------------------------------------------------------




# -*- coding: utf-8 -*-

#-------------------------------------------------------------------------------

# This file is part of Code_Saturne, a general-purpose CFD tool.
#
# Copyright (C) 1998-2021 EDF S.A.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.

#-------------------------------------------------------------------------------

"""
This module contains the creation of the dependancy graph between CASES.

This module contains the following classes:
- node_case
- dependency_graph
"""

#-------------------------------------------------------------------------------
# Standard modules import
#-------------------------------------------------------------------------------

import os, sys
import logging

#-------------------------------------------------------------------------------
# Application modules import
#-------------------------------------------------------------------------------


#-------------------------------------------------------------------------------
# log config
#-------------------------------------------------------------------------------

logging.basicConfig()
log = logging.getLogger(__file__)
#log.setLevel(logging.DEBUG)
log.setLevel(logging.NOTSET)

#-------------------------------------------------------------------------------
# class node_case
#-------------------------------------------------------------------------------

class node_case:

    def __init__(self, name, n_procs, n_iter, estim_wtime, tags, depends):
        """ initializes a Code_Saturne CASE
        """
        # name should be in the folling form: STUDY/CASE/run_id
        # ex: "81_HEAD_LOSSES_COMPONENT/CASE_BEND/mesh12"
        self.name        = name
        self.n_procs     = n_procs
        self.tags        = tags
        # TODO: verify that these parameters are well handled by smgr in general
        self.n_iter      = n_iter
        # estimated wall time for the submission of the case on cluster
        self.estim_wtime = estim_wtime
        # depends should be in the same format as name
        self.depends     = depends
        # level is modified in add_node
        self.level       = None

    def __str__(self):
        res  = '\nCase ' + self.name
        res += ' on ' + str(self.n_procs) + ' procs'
        res += ' on ' + str(self.n_iter) + ' iterations'
        res += ' with an estimated wall time of ' + str(self.estim_wtime)
        res += ' with tags ' + str(self.tags)
        res += ' with level ' + str(self.level)

        return res

#-------------------------------------------------------------------------------
# class dependency_graph
#-------------------------------------------------------------------------------

class dependency_graph(object):

    def __init__(self):
        """ Initializes a dependency graph object to an empty dictionary
        """
        self.graph_dict = {}
        self.add_node(node_case('root', n_procs=0, n_iter=0, estim_wtime=0,
                                tags=None, depends=None))

    def add_dependency(self, dependency):
        """ Defines dependency between two node_cases as an edge in the graph
        """
        (node1, node2) = dependency
        if node1 in self.graph_dict:
            self.graph_dict[node1].append(node2)
        else:
            self.graph_dict[node1] = [node2]

    def add_node(self, node):
        """ Add a node_case in the graph if not already there.
            Add a dependency when depends parameters is defined
        """
        if node not in self.graph_dict:
            self.graph_dict[node] = []

            root_node = self.root_node()
            if node is not root_node:
                if node.depends:
                    for neighbor in self.graph_dict:
                        if neighbor.name == node.depends:
                            # cases with dependency are level > 1 and connected to the dependency
                            self.add_dependency((node, neighbor))
                            node.level = neighbor.level + 1
                            break
                    if node.level is None:
                        msg = "Problem in graph construction : dependency " \
                              + node.depends + " is not found.\n"
                        sys.exit(msg)
                else:
                    # cases with no dependency are level 1 and connected to the root node
                    self.add_dependency((node, root_node))
                    node.level = 1
            else:
                # root node is level 0 and have no dependency
                node.level = 0

    def nodes(self):
        """ returns the cases of the dependency graph """
        return list(self.graph_dict.keys())

    def dependencies(self):
        """ returns the dependencies between the cases of the graph """
        dependencies = []
        for node in self.graph_dict:
            for neighbor in self.graph_dict[node]:
                if (neighbor, node) not in dependencies:
                    dependencies.append((node, neighbor))
        return dependencies

    def extract_sub_graph(self, level, n_procs):
        """ extracts a sub_graph based on level and n_procs criteria"""
        sub_graph = dependency_graph()
        for node in self.graph_dict:
            if node.level == level and int(node.n_procs) == n_procs:
                sub_graph.add_node(node)
        return sub_graph

    def root_node(self):
        """ returns the root node of the graph """
        for node in self.graph_dict:
            if node.name == 'root':
                return node

    def __str__(self):
        res = "\nList of cases: "
        for node in self.nodes():
            res += str(node) + " "
        res += "\nList of dependencies: "
        for dependency in self.dependencies():
            (node1, node2) = dependency
            res += '\n' + str(node1.name) + ' depends on ' + str(node2.name)
        return res

#-------------------------------------------------------------------------------




# -*- coding: utf-8 -*-

#-------------------------------------------------------------------------------

# This file is part of Code_Saturne, a general-purpose CFD tool.
#
# Copyright (C) 1998-2021 EDF S.A.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.

#-------------------------------------------------------------------------------

"""
This module contains the following classes:
- BoundaryConditionsNonCondensableView
"""

#-------------------------------------------------------------------------------
# Standard modules
#-------------------------------------------------------------------------------

import string, logging

#-------------------------------------------------------------------------------
# Third-party modules
#-------------------------------------------------------------------------------

from code_saturne.Base.QtCore    import *
from code_saturne.Base.QtGui     import *
from code_saturne.Base.QtWidgets import *

#-------------------------------------------------------------------------------
# Application modules import
#-------------------------------------------------------------------------------

from code_saturne.Pages.BoundaryConditionsNonCondensable import Ui_BoundaryConditionsNonCondensable

from code_saturne.model.Common import GuiParam
from code_saturne.Base.QtPage import DoubleValidator, ComboModel, from_qvariant

from code_saturne.model.NonCondensableModel import NonCondensableModel

#-------------------------------------------------------------------------------
# log config
#-------------------------------------------------------------------------------

logging.basicConfig()
log = logging.getLogger("BoundaryConditionsNonCondensableView")
log.setLevel(GuiParam.DEBUG)

#-------------------------------------------------------------------------------
# Main class
#-------------------------------------------------------------------------------

class BoundaryConditionsNonCondensableView(QWidget, Ui_BoundaryConditionsNonCondensable) :
    """
    Boundary condition for non condensable
    """
    def __init__(self, parent):
        """
        Constructor
        """
        QWidget.__init__(self, parent)

        Ui_BoundaryConditionsNonCondensable.__init__(self)
        self.setupUi(self)

        # Connections
        self.comboBoxNonCondensable.activated[str].connect(self.__slotChoiceNonCondensable)

        self.__NonCondensablemodel = ComboModel(self.comboBoxNonCondensable, 1, 1)

        self.lineEditNonCondensable.textChanged[str].connect(self.__slotNonCondensable)

        validatorNonCond = DoubleValidator(self.lineEditNonCondensable, min = 0.)
        validatorNonCond.setExclusiveMin(False)

        self.lineEditNonCondensable.setValidator(validatorNonCond)


    def setup(self, case, fieldId):
        """
        Setup the widget
        """
        self.case = case
        self.__boundary = None
        self.__currentField = fieldId


    def showWidget(self, boundary):
        """
        Show the widget
        """
        self.__boundary = boundary

        NonCondensableList = NonCondensableModel(self.case).getNonCondensableByFieldId(self.__currentField)

        if len(NonCondensableList) > 0 :
            for nb in range(len(self.__NonCondensablemodel.getItems())):
                self.__NonCondensablemodel.delItem(0)

            for var in NonCondensableList :
                name = NonCondensableModel(self.case).getNonCondLabel(var)
                self.__NonCondensablemodel.addItem(self.tr(name), var)

            self.__currentNonCondensable = NonCondensableList[0]
            self.__NonCondensablemodel.setItem(str_model=self.__currentNonCondensable)
            val = self.__boundary.getNonCondensableValue(self.__currentField, self.__currentNonCondensable)
            self.lineEditNonCondensable.setText(str(val))
            self.show()
        else :
            self.hideWidget()


    def hideWidget(self):
        """
        Hide the widget
        """
        self.hide()


    @pyqtSlot(str)
    def __slotChoiceNonCondensable(self, text):
        """
        INPUT choice of non condensable
        """
        self.__currentNonCondensable = self.__NonCondensablemodel.dicoV2M[str(text)]

        val = self.__boundary.getNonCondensableValue(self.__currentField, self.__currentNonCondensable)
        self.lineEditNonCondensable.setText(str(val))


    @pyqtSlot(str)
    def __slotNonCondensable(self, text):
        """
        INPUT non condensable value
        """
        if self.lineEditNonCondensable.validator().state == QValidator.Acceptable:
            value = from_qvariant(text, float)
            self.__boundary.setNonCondensableValue(self.__currentField, self.__currentNonCondensable, value)


#-------------------------------------------------------------------------------
# End
#-------------------------------------------------------------------------------




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#-------------------------------------------------------------------------------

# This file is part of Code_Saturne, a general-purpose CFD tool.
#
# Copyright (C) 1998-2021 EDF S.A.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.

#-------------------------------------------------------------------------------

try:
    import ConfigParser  # Python2
    configparser = ConfigParser
except Exception:
    import configparser  # Python3
import datetime
import fnmatch
import os
import os.path
import sys
import shutil
import stat

from code_saturne import cs_compile
from code_saturne import cs_xml_reader

from code_saturne.cs_exec_environment import run_command
from code_saturne.cs_exec_environment import enquote_arg, separate_args
from code_saturne.cs_exec_environment import get_ld_library_path_additions
from code_saturne.cs_exec_environment import source_syrthes_env

from code_saturne.cs_meg_to_c import meg_to_c_interpreter

#===============================================================================
# Utility functions
#===============================================================================

def any_to_str(arg):
    """Transform single values or lists to a whitespace-separated string"""

    s = ''

    if type(arg) == tuple or type(arg) == list:
        for e in arg:
            s += ' ' + str(e)
        return s[1:]

    else:
        return str(arg)

#-------------------------------------------------------------------------------

def make_clean_dir(path):
    """
    Create a directory, or remove files (not directories) in existing directory
    """

    if not os.path.isdir(path):
        os.mkdir(path)
    else:
        l = os.listdir(path)
        for f in l:
            os.remove(os.path.join(path, f))

#-------------------------------------------------------------------------------

class RunCaseError(Exception):
    """Base class for exception handling."""

    def __init__(self, *args):
        self.args = args

    def __str__(self):
        if len(self.args) == 1:
            return str(self.args[0])
        else:
            return str(self.args)

#   def __repr__(self):
#       return "%s(*%s)" % (self.__class__.__name__, repr(self.args))

#===============================================================================
# Classes
#===============================================================================

class base_domain:
    """
    Base class from which classes handling running case should inherit.
    """

    #---------------------------------------------------------------------------

    def __init__(self,
                 package,                 # main package
                 name = None,             # domain name
                 n_procs_weight = None,   # recommended number of processes
                 n_procs_min = 1,         # min. number of processes
                 n_procs_max = None):     # max. number of processes

        # Package specific information

        self.package = package

        # User functions

        self.user = {}

        # Names, directories, and files in case structure

        self.case_dir = None
        self.case_root_dir = None

        self.name = name # used for multiple domains only

        self.data_dir = None
        self.result_dir = None
        self.src_dir = None

        # Notebook and parametric  definitions and additional user arguments

        self.notebook = None
        self.parametric_args = None
        self.kw_args = None

        # Working directory and executable

        self.exec_dir = None
        self.solver_path = None

        # Execution and debugging options

        self.n_procs = n_procs_weight
        if not n_procs_min:
            n_procs_min = 1
        self.n_procs_min = max(1, n_procs_min)
        self.n_procs_max = n_procs_max

        if self.n_procs == None:
            self.n_procs = 1
        self.n_procs = max(self.n_procs, self.n_procs_min)
        if self.n_procs_max != None:
            self.n_procs = min(self.n_procs, self.n_procs_max)

        self.debug = None

        # Error reporting
        self.error = ''

        self.error_long = ''

    #---------------------------------------------------------------------------

    def set_case_dir(self, case_dir, staging_dir = None):

        # Names, directories, and files in case structure

        self.case_dir = case_dir
        self.case_root_dir = case_dir

        if self.name != None:
            self.case_dir = os.path.join(self.case_dir, self.name)
            self.case_root_dir = os.path.join(case_dir, self.name)

        self.data_dir = os.path.join(self.case_dir, 'DATA')
        self.result_dir = os.path.join(self.case_dir, 'RESU')
        self.src_dir = os.path.join(self.case_dir, 'SRC')

        # If computation is already staged, avoid reading data in upstream
        # case directories, as it may have changed since staging.

        if staging_dir:
            self.case_dir = staging_dir
            if self.name != None:
                self.case_dir = os.path.join(self.case_dir, self.name)
            self.data_dir = self.case_dir
            self.src_dir = None

    #---------------------------------------------------------------------------

    def set_exec_dir(self, exec_dir):

        if os.path.isabs(exec_dir):
            self.exec_dir = exec_dir
        else:
            self.exec_dir = os.path.join(self.case_dir, 'RESU', exec_dir)

        if self.name != None:
            self.exec_dir = os.path.join(self.exec_dir, self.name)

        if not os.path.isdir(self.exec_dir):
            os.makedirs(self.exec_dir)

    #---------------------------------------------------------------------------

    def set_result_dir(self, name, given_dir = None):
        """
        If suffix = true, add suffix to all names in result dir.
        Otherwise, create subdirectory
        """

        if given_dir == None:
            self.result_dir = os.path.join(self.case_dir, 'RESU', name)
        else:
            self.result_dir = given_dir

        if self.name != None:
            self.result_dir = os.path.join(self.result_dir, self.name)

        if not os.path.isdir(self.result_dir):
            os.makedirs(self.result_dir)

    #---------------------------------------------------------------------------

    def copy_data(self):
        """
        Copy base data to the execution directory.
        """

        return

    #---------------------------------------------------------------------------

    def init_staged_data(self):
        """
        Initialize staged data in the execution directory.
        """

        return

    #---------------------------------------------------------------------------

    def copy_result(self, name, purge=False):
        """
        Copy a file or directory to the results directory,
        optionally removing it from the source.
        """

        # Determine absolute source and destination names

        if os.path.isabs(name):
            src = name
            dest = os.path.join(self.result_dir, os.path.basename(name))
        else:
            src = os.path.join(self.exec_dir, name)
            dest = os.path.join(self.result_dir, name)

        # If source and destination are identical, return
        if src == dest:
            return

        # Copy single file

        if os.path.isfile(src):
            shutil.copy2(src, dest)
            if purge:
                os.remove(src)

        # Copy single directory (possibly recursive)
        # Unlike os.path.copytree, the destination directory
        # may already exist.

        elif os.path.isdir(src):

            if not os.path.isdir(dest):
                os.mkdir(dest)
            l = os.listdir(src)
            for f in l:
                f_src = os.path.join(src, f)
                f_dest = os.path.join(dest, f)
                if os.path.isfile(f_src):
                    shutil.copy2(f_src, f_dest)
                elif os.path.isdir(f_src):
                    self.copy_result(f_src, f_dest)

            if purge:
                if os.path.islink(src):
                    os.remove(f)
                else:
                    shutil.rmtree(src)

    #---------------------------------------------------------------------------

    def purge_result(self, name):
        """
        Remove a file or directory from execution directory.
        """

        # Determine absolute name

        if os.path.isabs(name):
            f = name
        else:
            f = os.path.join(self.exec_dir, name)

        # Remove file or directory

        if os.path.isfile(f) or os.path.islink(f):
            os.remove(f)

        elif os.path.isdir(f):
            shutil.rmtree(f)

    #---------------------------------------------------------------------------

    def get_n_procs(self):
        """
        Returns an array (list) containing the current number of processes
        associated with a solver stage followed by the minimum and maximum
        number of processes.
        """

        return [self.n_procs, self.n_procs_min, self.n_procs_max]

    #---------------------------------------------------------------------------

    def set_n_procs(self, n_procs):
        """
        Assign a number of processes to a solver stage.
        """

        self.n_procs = n_procs

    #---------------------------------------------------------------------------

    def solver_command(self, **kw):
        """
        Returns a tuple indicating the solver's working directory,
        executable path, and associated command-line arguments.
        """

        return enquote_arg(self.exec_dir), enquote_arg(self.solver_path), ''

    #---------------------------------------------------------------------------

    def summary_info(self, s):
        """
        Output summary data into file s
        """

        if self.name:
            name = self.name
            exec_dir = os.path.join(self.exec_dir, name)
            result_dir = os.path.join(self.result_dir, name)
        else:
            name = os.path.basename(self.case_dir)
            exec_dir = self.exec_dir
            result_dir = self.result_dir

        s.write('  Case           : ' + name + '\n')
        s.write('    directory    : ' + self.case_dir + '\n')
        s.write('    results dir. : ' + self.result_dir + '\n')
        if exec_dir != result_dir:
            s.write('    exec. dir.   : ' + self.exec_dir + '\n')

    #---------------------------------------------------------------------------

    def debug_wrapper_args(self, debug_cmd):
        """
        Additional arguments to use debug wrapper
        """
        debug_args = ''
        python_exec = self.package.config.python
        if os.path.isfile(python_exec) or os.path.islink(python_exec):
            debug_args += python_exec + ' '
        else:
            debug_args += 'python '
        cs_pkg_dir = self.package.get_dir('pkgpythondir')
        if self.package.name != 'code_saturne':
            cs_pkg_dir = os.path.join(cs_pkg_dir, '../code_saturne')
        dbg_wrapper_path = os.path.join(cs_pkg_dir, 'cs_debug_wrapper.py')
        debug_args += dbg_wrapper_path + ' '
        for a in separate_args(debug_cmd):
            debug_args += enquote_arg(a) + ' '
        return debug_args

#-------------------------------------------------------------------------------

class domain(base_domain):
    """Handle running case."""

    #---------------------------------------------------------------------------

    def __init__(self,
                 package,                     # main package
                 package_compute = None,      # package for compute environment
                 name = None,                 # domain name
                 n_procs_weight = None,       # recommended number of processes
                 n_procs_min = None,          # min. number of processes
                 n_procs_max = None,          # max. number of processes
                 logging_args = None,         # command-line options for logging
                 param = None,                # XML parameters file
                 prefix = None,               # installation prefix
                 adaptation = None):          # HOMARD adaptation script

        base_domain.__init__(self, package,
                             name,
                             n_procs_weight,
                             n_procs_min,
                             n_procs_max)

        # Compute package if different from front-end

        if package_compute:
            self.package_compute = package_compute
        else:
            self.package_compute = self.package

        # Directories, and files in case structure

        self.restart_input = None
        self.restart_mesh_input = None
        self.mesh_input = None
        self.partition_input = None

        # Default executable

        self.solver_path = os.path.join(self.package_compute.get_dir("pkglibexecdir"),
                                        self.package.solver)

        # Preprocessor options

        self.mesh_dir = None
        self.meshes = None

        # Solver options

        self.preprocess_on_restart = False
        self.exec_solver = True

        if param:
            self.param = os.path.basename(param)
        else:
            self.param = None

        self.logging_args = logging_args
        self.solver_args = None

        self.debug = None

        # Additional data

        self.prefix = prefix

        self.compile_cflags = None
        self.compile_cxxflags = None
        self.compile_fcflags = None
        self.compile_libs = None

        # Adaptation using HOMARD

        self.adaptation = adaptation

        # MEG expression generator
        self.mci = None

    #---------------------------------------------------------------------------

    def __set_case_parameters__(self, update_xml=False):

        # We may now import user python script functions if present.

        self.user_locals = None

        user_scripts = os.path.join(self.exec_dir, 'cs_user_scripts.py')
        if os.path.isfile(user_scripts):
            try:
                exec(compile(open(user_scripts).read(), user_scripts, 'exec'),
                     locals(),
                     locals())
                self.user_locals = locals()
            except Exception:
                execfile(user_scripts, locals(), locals())
                self.user_locals = locals()

        # We may now parse the optional XML parameter file
        # now that its path may be built and checked.

        setup_path = os.path.join(self.exec_dir, "setup.xml")
        if os.path.isfile(setup_path):

            # Ensure XML file is up to date as a precaution,
            # and filter it in case of parametric or notebook arguments.

            case = self.__xml_case_initialize__(setup_path, apply_filters=True)
            case['xmlfile'] = setup_path

            P = cs_xml_reader.Parser(doc=case.doc)
            params = P.getParams()
            for k in list(params.keys()):
                self.__dict__[k] = params[k]

            case.xmlSaveDocument()

            if params['xml_root_name'] == 'NEPTUNE_CFD_GUI':
                solver_dir = self.package_compute.get_dir("pkglibexecdir")
                solver_name = "nc_solver" + self.package_compute.config.exeext
                self.solver_path = os.path.join(solver_dir, solver_name)

            self.param = "setup.xml"

        else:
            msg  = ('Remark:\n'
                    '  No setup.xml file was provided in the DATA folder.\n'
                    '  Default settings will be used.\n')
            print(msg, file = sys.stderr)

        # Now override or complete data from the XML file.

        if self.user_locals:
            m = 'define_domain_parameters'
            if m in self.user_locals.keys():
                eval(m + '(self)', globals(), self.user_locals)
                del self.user_locals[m]

        # Finally, ensure some fields are of the required types

        if type(self.meshes) != list:
            self.meshes = [self.meshes,]

    #---------------------------------------------------------------------------

    def __set_auto_restart__(self):
        """
        Select latest valid checkpoint directory for restart, based on name
        """

        self.restart_input = None

        from code_saturne.cs_exec_environment import get_command_output

        results_dir = os.path.abspath(os.path.join(self.result_dir, '..'))
        results = os.listdir(results_dir)
        results.sort(reverse=True)
        for r in results:
            m = os.path.join(results_dir, r, 'checkpoint', 'main.csc')
            if not os.path.isfile(m):
                m = os.path.join(results_dir, r, 'checkpoint', 'main')
            if os.path.isfile(m):
                try:
                    cmd = self.package.get_io_dump()
                    cmd += ' --location 0 ' + m
                    res = get_command_output(cmd)
                except Exception:
                    print('checkpoint of result: ' + r + ' does not seem usable')
                    continue
                self.restart_input = os.path.join(results_dir, r, 'checkpoint')
                break

        return

    #---------------------------------------------------------------------------

    def __xml_case_initialize__(self, path, apply_filters=False):
        """
        Build XML case object
        """

        from code_saturne.model.XMLengine import Case
        from code_saturne.model.SolutionDomainModel import getRunType
        case = Case(package=self.package, file_name=path)
        case['xmlfile'] = path
        case.xmlCleanAllBlank(case.xmlRootNode())
        preprocess_only = (getRunType(case) != 'standard')
        module_name = case.module_name()
        if module_name == 'code_saturne':
            from code_saturne.model.XMLinitialize import XMLinit
            XMLinit(case).initialize(preprocess_only)
        elif module_name == 'neptune_cfd':
            from code_saturne.model.XMLinitializeNeptune import XMLinitNeptune
            XMLinitNeptune(case).initialize(preprocess_only)

        if not apply_filters:
            return case

        # Apply changes defined through notebook or parametric options.

        if self.parametric_args:
            from code_saturne import cs_parametric_setup
            cs_parametric_setup.update_case_setup(case, self.parametric_args,
                                                  pkg=self.package)

        if self.notebook:
            from code_saturne.model.NotebookModel import NotebookModel
            notebookModel = NotebookModel(case)
            nbk_vars = notebookModel.getVarNameList()
            n_warnings = 0
            for k in self.notebook.keys():
                if k in nbk_vars:
                    vs = self.notebook[k]
                    v = None
                    try:
                        v = float(vs)
                    except Exception:
                        fmt = ("Warning: notebook variable '{0}'='{1}'"
                               " is not a real number.")
                        msg = fmt.format(k, vs)
                        print(msg, file = sys.stderr)
                        n_warnings += 1
                    if v != None:
                        notebookModel.setVariableValue(val=v, var=k)

                else:
                    fmt = ('Warning: {0} is not a known notebook variable.')
                    msg = fmt.format(k)
                    print(msg, file = sys.stderr)
                    n_warnings += 1
            if n_warnings > 0:
                print(file = sys.stderr)

        return case

    #---------------------------------------------------------------------------

    def for_domain_str(self):

        if self.name == None:
            return ''
        else:
            return 'for domain ' + str(self.name)

    #---------------------------------------------------------------------------

    def read_parameter_file(self, param):
        """
        Parse the optional XML parameter file.
        """

        if param == None:
            if os.path.isfile(os.path.join(self.data_dir, 'setup.xml')):
                param = 'setup.xml'

        if param != None:
            version_str = '2.0'
            P = cs_xml_reader.Parser(os.path.join(self.data_dir, param),
                                     version_str = version_str)
            params = P.getParams()
            for k in list(params.keys()):
                self.__dict__[k] = params[k]

            self.param = param

    #---------------------------------------------------------------------------

    def copy_data(self):
        """
        Copy base data to the execution directory
        """

        err_str = ""

        # Create the src folder if there are files to compile in source path

        src_files = []
        if os.path.exists(self.src_dir):
            src_files = cs_compile.files_to_compile(self.src_dir)
        if len(src_files) > 0:
            exec_src = os.path.join(self.exec_dir, 'src')

            make_clean_dir(exec_src)

            # Add header files to list so as not to forget to copy them
            dir_files = os.listdir(self.src_dir)
            src_files = src_files + (  fnmatch.filter(dir_files, '*.h')
                                     + fnmatch.filter(dir_files, '*.hxx')
                                     + fnmatch.filter(dir_files, '*.hpp'))

            # Copy source files to execution directory
            for f in src_files:
                src_file = os.path.join(self.src_dir, f)
                dest_file = os.path.join(exec_src, f)
                shutil.copy2(src_file, dest_file)

        # Copy data files

        dir_files = os.listdir(self.data_dir)

        if self.package.name in dir_files:
            dir_files.remove(self.package.name)

        for f in dir_files:
            src = os.path.join(self.data_dir, f)
            if os.path.isfile(src):
                shutil.copy2(src, os.path.join(self.exec_dir, f))

        # Fixed parameter name

        setup_path = os.path.join(self.exec_dir, "setup.xml")

        if self.param != None:
            param_base = os.path.basename(self.param)
            if param_base != "setup.xml":
                src_path = os.path.join(self.exec_dir, param_base)
                if os.path.isfile(setup_path):
                    self.purge_result(setup_path) # in case of previous run here
                    fmt = ('Warning:\n'
                           '  Both {0} and {1} exist in\n'
                           '    {2}.\n'
                           '  {0} will be used for the computation.\n'
                           '  Be aware that to follow best practices '
                           'only one of the two should be present.\n\n')
                    msg = fmt.format(os.path.basename(self.param),
                                     os.path.basename(setup_path),
                                     self.data_dir)
                    print(msg, file = sys.stderr)
                try:
                    if os.path.islink(setup_path):
                        os.remove(setup_path)
                    os.symlink(self.param, setup_path)
                except Exception:
                    shutil.copy2(src_path, setup_path)

        # Now set parameters

        self.__set_case_parameters__(update_xml=True)

    #---------------------------------------------------------------------------

    def init_staged_data(self):
        """
        Initialize staged data in the execution directory.
        """

        # Now set parameters

        self.__set_case_parameters__(update_xml=False)

        # Ensure correct executable is used.

        exec_src = os.path.join(self.exec_dir, 'src')
        if os.path.isdir(exec_src):
            if len(cs_compile.files_to_compile(exec_src)) > 0:
                solver_name = os.path.basename(self.solver_path)
                self.solver_path = os.path.join('.', solver_name)

    #---------------------------------------------------------------------------

    def symlink(self, target, link=None, check_type=None):
        """
        Create a symbolic link to a file, or copy it if links are
        not possible
        """

        if target == None and link == None:
            return
        elif target == None:
            err_str = 'No target for link: ' + link
            raise RunCaseError(err_str)
        elif link == None:
            if self.exec_dir != None:
                link = os.path.join(self.exec_dir,
                                    os.path.basename(target))
            else:
                err_str = 'No path name given for link to: ' + target
                raise RunCaseError(err_str)

        if not os.path.exists(target):
            err_str = 'File: ' + target + ' does not exist.'
            raise RunCaseError(err_str)

        elif check_type == 'file':
            if not os.path.isfile(target):
                err_str = target + ' is not a regular file.'
                raise RunCaseError(err_str)

        elif check_type == 'dir':
            if not os.path.isdir(target):
                err_str = target + ' is not a directory.'
                raise RunCaseError(err_str)

        try:
            os.symlink(target, link)
        except AttributeError:
            if not os.path.isdir(target):
                shutil.copy2(target, link)
            else:
                if os.path.isdir(link):
                    shutil.rmtree(link)
                shutil.copytree(target, link)

    #---------------------------------------------------------------------------

    def needs_compile(self):
        """
        Compile and link user subroutines if necessary
        """
        # Check if there are files to compile in source path

        needs_comp = False

        exec_src = os.path.join(self.exec_dir, 'src')
        if os.path.isdir(exec_src):
            src_files = cs_compile.files_to_compile(exec_src)
            if self.exec_solver and len(src_files) > 0:
                needs_comp = True

        setup_path = os.path.join(self.exec_dir, "setup.xml")
        if os.path.isfile(setup_path):
            fp = os.path.join(self.exec_dir, self.param)
            case = self.__xml_case_initialize__(fp)

            # Do not call case.xmlSaveDocument() to avoid side effects in case
            # directory; is not required as meg_to_c_interpreter works from
            # case in memory

            module_name = case.module_name()
            self.mci = meg_to_c_interpreter(case,
                                            module_name=module_name,
                                            wdir = os.path.join(self.exec_dir, 'src'))

            if self.mci.has_meg_code():
                needs_comp = True
            else:
                self.mci = None

        return needs_comp

    #---------------------------------------------------------------------------

    def compile_and_link(self):
        """
        Compile and link user subroutines if necessary
        """

        src_files = []

        # Check if there are files to compile in source path
        # or if MEG functions need to be generated
        exec_src = os.path.join(self.exec_dir, 'src')
        if os.path.isdir(exec_src):
            src_files = cs_compile.files_to_compile(exec_src)

        if len(src_files) > 0 or self.mci != None:

            exec_src = os.path.join(self.exec_dir, 'src')

            # Create the src folder if not done yet
            if not os.path.isdir(exec_src):
                os.mkdir(exec_src)

            if len(src_files) > 0:
                # Add header files to list so as not to forget to copy them
                dir_files = os.listdir(self.src_dir)
                src_files = src_files + (  fnmatch.filter(dir_files, '*.h')
                                         + fnmatch.filter(dir_files, '*.hxx')
                                         + fnmatch.filter(dir_files, '*.hpp'))

            if self.mci != None:
                mci_state = self.mci.save_all_functions()
                if mci_state['state'] == -1:
                    self.error = 'compile or link'
                    self.error_long = ' missing mathematical expressions:\n\n'
                    for i, eme in enumerate(mci_state['exps']):
                        self.error_long += " (%d/%d) %s is not provided for %s for zone %s\n" % (i+1, mci_state['nexps'], eme['func'], eme['var'], eme['zone'])

                    return
                elif mci_state['state'] == 2:
                    self.error = 'saving MEG generated sources'
                    self.error_long = ' Incorrect directory ?'

            log_name = os.path.join(self.exec_dir, 'compile.log')
            log = open(log_name, 'w')

            solver_name = os.path.basename(self.solver_path)

            retval = cs_compile.compile_and_link(self.package_compute,
                                                 solver_name,
                                                 exec_src,
                                                 self.exec_dir,
                                                 self.compile_cflags,
                                                 self.compile_cxxflags,
                                                 self.compile_fcflags,
                                                 self.compile_libs,
                                                 keep_going=True,
                                                 stdout=log,
                                                 stderr=log)

            log.close()

            if retval == 0:
                solver_dir = '.'
                self.solver_path = os.path.join(solver_dir, solver_name)
            else:
                # In case of error, copy source to results directory now,
                # as no calculation is possible, then raise exception
                for f in ['src', 'compile.log']:
                    self.copy_result(f)
                self.error = 'compile or link'

    #---------------------------------------------------------------------------

    def prepare_data(self):
        """
        Copy data to the execution directory
        """

        err_str = ""

        if not self.exec_solver:
            return

        # Call user script if necessary

        if self.user_locals:
            m = 'domain_prepare_data_add'
            if m in self.user_locals.keys():
                eval(m + '(self)', globals(), self.user_locals)
                del self.user_locals[m]

        # Restart files

        # Handle automatic case first

        ignore_checkpoint = False
        if self.solver_args:
            for a in ('--preprocess', '--quality', '-q'):
                if a in self.solver_args:
                    ignore_checkpoint = True

        if not ignore_checkpoint:
            if self.restart_input == '*':
                self.__set_auto_restart__()

            if self.restart_input != None:

                restart_input =  os.path.expanduser(self.restart_input)
                if not os.path.isabs(restart_input):
                    restart_input = os.path.join(self.case_dir, restart_input)

                if not os.path.exists(restart_input):
                    err_str += restart_input + ' does not exist.\n\n'
                elif not os.path.isdir(restart_input):
                    err_str += restart_input + ' is not a directory.\n\n.'
                else:
                    self.symlink(restart_input,
                                 os.path.join(self.exec_dir, 'restart'))

                print(' Restart from ' + self.restart_input + '\n')

            if self.restart_mesh_input != None and err_str == '':

                restart_mesh_input =  os.path.expanduser(self.restart_mesh_input)
                if not os.path.isabs(restart_mesh_input):
                    restart_mesh_input = os.path.join(self.case_dir,
                                                      restart_mesh_input)

                if not os.path.exists(restart_mesh_input):
                    err_str += restart_mesh_input + ' does not exist.\n\n'
                elif not os.path.isfile(restart_mesh_input):
                    err_str += restart_mesh_input + ' is not a file.\n\n.'
                else:
                    self.symlink(restart_mesh_input,
                                 os.path.join(self.exec_dir, 'restart_mesh_input'))

                print(' Restart mesh ' + self.restart_mesh_input + '\n')

        # Mesh input file

        restart_input_mesh = None
        if self.restart_input:
            restart_input_mesh = os.path.join(self.exec_dir, 'restart', 'mesh_input.csm')
            if not os.path.exists(restart_input_mesh):
                restart_input_mesh = None

        if restart_input_mesh is None or self.preprocess_on_restart \
           or self.restart_mesh_input:
            if self.mesh_input:
                mesh_input = os.path.expanduser(self.mesh_input)
                if not os.path.isabs(mesh_input):
                    mesh_input = os.path.join(self.case_dir, mesh_input)

                # Differentiate between a folder and file, since we now
                # have a file extension
                if os.path.isdir(mesh_input):
                    link_path = os.path.join(self.exec_dir, 'mesh_input')
                else:
                    link_path = os.path.join(self.exec_dir, 'mesh_input.csm')

                self.purge_result(link_path) # in case of previous run here
                self.symlink(mesh_input, link_path)
        else:
            # use mesh from restart, with no symbolic link
            self.mesh_input = restart_input_mesh

        # Partition input files

        if self.partition_input != None and not err_str:

            partition_input = os.path.expanduser(self.partition_input)
            if not os.path.isabs(partition_input):
                partition_input = os.path.join(self.case_dir, partition_input)

            if os.path.exists(partition_input):

                if not os.path.isdir(partition_input):
                    err_str += partition_input + ' is not a directory.\n\n'
                else:
                    self.symlink(partition_input,
                                 os.path.join(self.exec_dir, 'partition_input'))

        if len(err_str) > 0:
            self.error = 'data preparation'
            sys.stderr.write(err_str)
        else:
            try:
                link_path = os.path.join(self.exec_dir, "listing")
                os.symlink("run_solver.log", link_path)
            except Exception:
                pass

    #---------------------------------------------------------------------------

    def preprocess(self):
        """
        Runs the preprocessor in the execution directory
        """

        if self.mesh_input:
            return

        # Check if cartesian mesh is to be used
        if self.param != None:
            from code_saturne.model.XMLengine import Case
            from code_saturne.model.SolutionDomainModel import getMeshOriginType

            fp = os.path.join(self.exec_dir, self.param)
            case = Case(package=self.package, file_name=fp)
            case['xmlfile'] = fp
            case.xmlCleanAllBlank(case.xmlRootNode())

            if getMeshOriginType(case) == 'mesh_cartesian':
                return

        # If no mesh is provided return, since user can define mesh_input
        # using 'cs_user_mesh_input' user function.
        if len(self.meshes) == 1 and self.meshes[0] == None:
            return

        # Study directory
        study_dir = os.path.split(self.case_root_dir)[0]

        # User config file
        u_cfg = configparser.ConfigParser()
        u_cfg.read(self.package.get_user_configfile())

        # Global config file
        g_cfg = configparser.ConfigParser()
        g_cfg.read(self.package.get_global_configfile())

        # A mesh can be found in different mesh database directories
        # (case, study, user, global -- in this order)
        mesh_dirs = []
        if self.mesh_dir is not None:
            mesh_dir = os.path.expanduser(self.mesh_dir)
            if not os.path.isabs(mesh_dir):
                mesh_dir = os.path.join(self.case_root_dir, mesh_dir)
            mesh_dirs.append(mesh_dir)
        if os.path.isdir(os.path.join(study_dir, 'MESH')):
            mesh_dirs.append(os.path.join(study_dir, 'MESH'))
        if u_cfg.has_option('run', 'meshdir'):
            add_path = u_cfg.get('run', 'meshdir').split(':')
            for d in add_path:
                mesh_dirs.append(d)
        if g_cfg.has_option('run', 'meshdir'):
            add_path = g_cfg.get('run', 'meshdir').split(':')
            for d in add_path:
                mesh_dirs.append(d)

        # Switch to execution directory

        cur_dir = os.path.realpath(os.getcwd())
        if cur_dir != self.exec_dir:
            os.chdir(self.exec_dir)

        mesh_id = None

        if len(self.meshes) > 1:
            mesh_id = 0
            destdir = 'mesh_input'
            make_clean_dir(destdir)

        # Set environment

        ld_library_path_save = None
        add_lib_dirs = get_ld_library_path_additions(self.package)
        if add_lib_dirs:
            ld_library_path_save = os.getenv('LD_LIBRARY_PATH')
            ld_library_path = ""
            for d in add_lib_dirs:
                ld_library_path += d + ':'
            ld_library_path += ld_library_path_save
            os.environ['LD_LIBRARY_PATH'] = ld_library_path

        # Run once per mesh

        for m in self.meshes:

            # Get absolute mesh paths

            if m is None:
                err_str = 'Preprocessing stage required but no mesh is given'
                raise RunCaseError(err_str)

            if (type(m) == tuple):
                m0 = m[0]
            else:
                m0 = m

            m0 = os.path.expanduser(m0)

            mesh_path = m0
            if (not os.path.isabs(m0)) and len(mesh_dirs) > 0:
                for mesh_dir in mesh_dirs:
                    mesh_path = os.path.join(mesh_dir, m0)
                    if os.path.isfile(mesh_path):
                        break

            if not os.path.isfile(mesh_path):
                err_str = 'Mesh file ' + m0 + ' not found'
                if not (os.path.isabs(mesh_path) or mesh_dirs):
                    err_str += '(no mesh directory given)'
                raise RunCaseError(err_str)

            # Build command

            # Generate output mesh name
            if (mesh_id != None):
                mesh_id += 1
                _outputmesh = os.path.join('mesh_input',
                                           'mesh_%02d.csm' % (mesh_id))
            else:
                _outputmesh = 'mesh_input.csm'

            # Check if preprocessor is needed or not
            if mesh_path[-4:] == ".csm":
                # code_saturne mesh, no need to run preprocessor
                self.symlink(mesh_path,
                             os.path.join(self.exec_dir, _outputmesh))

                retcode = 0
            else:
                # run preprocessor if needed

                cmd = [self.package.get_preprocessor()]

                if (type(m) == tuple):
                    for opt in m[1:]:
                        cmd.append(opt)

                cmd = cmd + ['--out', _outputmesh]
                if (mesh_id != None):
                    cmd = cmd + ['--log', 'preprocessor_%02d.log' % (mesh_id)]
                    cmd = cmd + ['--case', 'preprocessor_%02d' % (mesh_id)]
                else:
                    cmd = cmd + ['--log']

                cmd.append(mesh_path)

                # Run command
                retcode = run_command(cmd, pkg=self.package)

            if retcode != 0:
                err_str = \
                    'Error running the preprocessor.\n' \
                    'Check the preprocessor.log file for details.\n\n'
                sys.stderr.write(err_str)

                self.exec_solver = False

                self.error = 'preprocess'

                break

        # Restore environment

        if ld_library_path_save:
            os.environ['LD_LIBRARY_PATH'] = ld_library_path_save

        # Revert to initial directory

        if cur_dir != self.exec_dir:
            os.chdir(cur_dir)

        return retcode

    #---------------------------------------------------------------------------

    def solver_command(self, **kw):
        """
        Returns a tuple indicating the solver's working directory,
        executable path, and associated command-line arguments.
        """

        wd = enquote_arg(self.exec_dir)              # Working directory
        exec_path = enquote_arg(self.solver_path)    # Executable

        # Build kernel command-line arguments

        args = ''

        if self.logging_args != None:
            args += ' ' + self.logging_args

        if self.solver_args != None:
            args += ' ' + self.solver_args

        if self.package_compute.config.features['mpi'] == 'yes':
            if self.name != None:
                args += ' --mpi --app-name ' + enquote_arg(self.name)
            elif self.n_procs > 1:
                args += ' --mpi'

        # Adjust for debugger if used

        if self.debug != None:
            args = enquote_arg(self.solver_path) + ' ' + args
            exec_path = self.debug_wrapper_args(self.debug)

        return wd, exec_path, args

    #---------------------------------------------------------------------------

    def copy_results(self):
        """
        Retrieve solver results from the execution directory
        """

        # Call user script

        if self.user_locals:
            m = 'domain_copy_results_add'
            if m in self.user_locals.keys():
                eval(m + '(self)', globals(), self.user_locals)
                del self.user_locals[m]

        # Determine all files present in execution directory

        dir_files = os.listdir(self.exec_dir)

        # Only purge if we have no errors, to allow for future debugging.

        valid_dir = False
        purge = True

        if self.error != '':
            purge = False

        # Determine patterns from previous stages to ignore or possibly remove

        purge_list = []

        if not self.mesh_input and self.exec_solver:
            for f in ['mesh_input', 'mesh_input.csm']:
                if f in dir_files:
                    purge_list.append(f)

        for f in ['partition_input']:
            if f in dir_files:
                purge_list.append(f)

        # Determine files from this stage to ignore or to possibly remove

        solver_name = os.path.basename(self.solver_path)
        for f in [solver_name, self.package.runsolver]:
            if f in dir_files:
                purge_list.append(f)

        for f in purge_list:
            dir_files.remove(f)
            if purge:
                self.purge_result(f)

        if len(purge_list) > 0:
            valid_dir = True

        # Copy user sources, compile log, and xml file if present

        for f in ['src', 'compile.log', self.param]:
            if f in dir_files:
                valid_dir = True
                self.copy_result(f, purge)
                dir_files.remove(f)

        # Copy log files

        log_files = fnmatch.filter(dir_files, 'listing*')
        log_files.extend(fnmatch.filter(dir_files, '*.log'))
        log_files.extend(fnmatch.filter(dir_files, 'error*'))

        for f in log_files:
            self.copy_result(f, purge)
            dir_files.remove(f)

        if (len(log_files) > 0):
            valid_dir = True

        # Copy checkpoint files (in case of full disk, copying them
        # before other large data such as postprocessing output
        # increases chances of being able to continue).

        cpt = 'checkpoint'
        if cpt in dir_files:
            valid_dir = True
            self.copy_result(cpt, purge)
            dir_files.remove(cpt)

        # Now copy all other files

        if not valid_dir:
            return

        for f in dir_files:
            self.copy_result(f, purge)

    #---------------------------------------------------------------------------

    def summary_info(self, s):
        """
        Output summary data into file s
        """

        base_domain.summary_info(self, s)

        if not self.mesh_input:
            p = self.package.get_preprocessor()
            s.write('    preprocessor : ' + any_to_str(p) + '\n')

        if self.exec_solver:
            s.write('    solver       : ' + self.solver_path + '\n')

#-------------------------------------------------------------------------------

# SYRTHES coupling

class syrthes_domain(base_domain):

    def __init__(self,
                 package,
                 cmd_line = None,     # Command line to define optional syrthes4 behavior
                 name = None,
                 param = 'syrthes.data',
                 log_file = None,
                 n_procs_weight = None,
                 n_procs_min = 1,
                 n_procs_max = None,
                 n_procs_radiation = None,
                 verbose = True):

        base_domain.__init__(self,
                             package,
                             name,
                             n_procs_weight,
                             n_procs_min,
                             n_procs_max)

        self.n_procs_radiation = n_procs_radiation

        self.n_procs_ref = None
        self.n_procs_radiation = None

        # Additional parameters for Code_Saturne/SYRTHES coupling
        # Directories, and files in case structure

        self.cmd_line = cmd_line
        self.param = param

        self.logfile = log_file
        if self.logfile == None:
            self.logfile = 'syrthes.log'

        self.case_dir = None
        self.exec_dir = None
        self.data_dir = None
        self.src_dir = None
        self.result_dir = None
        self.echo_comm = None

        self.exec_solver = True

        # Generation of SYRTHES case deferred until we know how
        # many processors are really required

        self.syrthes_case = None

        # Determine environment which should be used for Syrthes
        # operations

        ld_library_path_save = os.getenv('LD_LIBRARY_PATH')

        source_syrthes_env(self.package, verbose)

        self.ld_library_path = os.getenv('LD_LIBRARY_PATH')

        self.__ld_library_path_restore__(ld_library_path_save)

    #---------------------------------------------------------------------------

    def __ld_library_path_restore__(self, ld_library_path_save):

        if ld_library_path_save:
            os.environ['LD_LIBRARY_PATH'] = ld_library_path_save
        else:
            try:
                os.environ.pop('LD_LIBRARY_PATH')
            except Exception:
                pass

    #---------------------------------------------------------------------------

    def set_case_dir(self, case_dir, staging_dir = None):

        base_domain.set_case_dir(self, case_dir, staging_dir)

        # Names, directories, and files in case structure

        self.data_dir = self.case_dir
        self.src_dir = self.case_dir

    #---------------------------------------------------------------------------

    def set_exec_dir(self, exec_dir):

        if os.path.isabs(exec_dir):
            self.exec_dir = exec_dir
        else:
            self.exec_dir = os.path.join(self.case_dir, 'RESU', exec_dir)

        self.exec_dir = os.path.join(self.exec_dir, self.name)

        if not os.path.isdir(self.exec_dir):
            os.mkdir(self.exec_dir)

    #---------------------------------------------------------------------------

    def set_result_dir(self, name, given_dir = None):

        if given_dir == None:
            self.result_dir = os.path.join(self.result_dir,
                                           'RESU_' + self.name,
                                           name)
        else:
            self.result_dir = os.path.join(given_dir, self.name)

        if not os.path.isdir(self.result_dir):
            os.makedirs(self.result_dir)

    #---------------------------------------------------------------------------

    def solver_command(self, **kw):
        """
        Returns a tuple indicating SYRTHES's working directory,
        executable path, and associated command-line arguments.
        """

        wd = enquote_arg(self.exec_dir)              # Working directory
        exec_path = enquote_arg(self.solver_path)    # Executable

        # Build kernel command-line arguments

        args = ''

        args += ' -d ' + enquote_arg(self.syrthes_case.data_file)
        args += ' -n ' + str(self.syrthes_case.n_procs)

        if self.syrthes_case.n_procs_ray > 0:
            args += ' -r ' + str(self.n_procs_ray)

        args += ' --name ' + enquote_arg(self.name)

        # Output to a logfile
        args += ' --log ' + enquote_arg(self.logfile)

        # Adjust for Debug if used

        if self.debug != None:
            args = enquote_arg(self.solver_path) + ' ' + args
            exec_path = self.debug_wrapper_args(self.debug)

        return wd, exec_path, args

    #---------------------------------------------------------------------------

    def __init_syrthes_case__(self):
        """
        Build or rebuild syrthes object
        """

        # Now try to read additional data set by Syrthes GUI

        part_tool_name = None

        f = open(os.path.join(self.case_dir, self.param), 'r')
        lines = f.readlines()
        del(f)
        for l in lines:
            if l[0] == '/':
                i = l.find('DOMAIN_POS=')
                if i > -1:
                    try:
                        id = int(l[i + 11:].strip())
                        if id == 0:
                            part_tool_name = 'scotch'
                        elif id == 1:
                            part_tool_name = 'metis'
                    except Exception:
                        pass
                i = l.find('LISTING=')
                if i > -1:
                    logfile = l[i + 8:].strip()
                    if logfile:
                        self.logfile = logfile
                i = l.find('NBPROC_COND=')
                if i > -1:
                     try:
                        self.n_procs_ref = int(l[i + 12:].strip())
                     except Exception:
                        pass
                i = l.find('NBPROC_RAD=')
                if i > -1:
                     try:
                        self.n_procs_radiation_ref = int(l[i + 11:].strip())
                     except Exception:
                        pass

        # Build command-line arguments

        args = '-d ' + os.path.join(self.case_root_dir, self.param)
        args += ' --name ' + self.name

        if self.n_procs != None and self.n_procs != 1:
            args += ' -n ' + str(self.n_procs)

        if self.n_procs_radiation != None:
            if self.n_procs_radiation > 0:
                args += ' -r ' + str(self.n_procs_radiation)

        if part_tool_name:
            args += ' -t ' + part_tool_name

        if self.data_dir != None:
            args += ' --data-dir ' + str(self.data_dir)

        if self.src_dir != None:
            args += ' --src-dir ' + str(self.src_dir)

        if self.exec_dir != None:
            args += ' --exec-dir ' + str(self.exec_dir)

        if self.cmd_line != None and len(self.cmd_line) > 0:
            args += ' ' + self.cmd_line

        # Define syrthes case structure

        import syrthes
        self.syrthes_case = syrthes.process_cmd_line(args.split())

        if self.syrthes_case.logfile == None:
            self.syrthes_case.set_logfile(self.logfile)
        else:
            self.logfile = self.syrthes_case.logfile

        # Read data file and store parameters

        self.syrthes_case.read_data_file()

    #---------------------------------------------------------------------------

    def prepare_data(self):
        """
        Fill SYRTHES domain structure
        Copy data to the execution directory
        Compile and link syrthes executable
        """

        # Build SYRTHES case; at this stage, the number of processes
        # may not be known, but this is not an issue here: for data
        # preparation, SYRTHES only uses the number of processes to
        # determine whether compilation should use MPI, and this is
        # always true anyways in the coupled case.

        ld_library_path_save = os.getenv('LD_LIBRARY_PATH')
        os.environ['LD_LIBRARY_PATH'] = self.ld_library_path

        self.__init_syrthes_case__()

        # Build exec_srcdir

        exec_srcdir = os.path.join(self.exec_dir, 'src')
        make_clean_dir(exec_srcdir)

        # Preparation of the execution directory and compile and link executable

        compile_logname = os.path.join(self.exec_dir, 'compile.log')

        retval = self.syrthes_case.prepare_run(exec_srcdir, compile_logname)

        self.copy_result(compile_logname)

        if retval != 0:
            err_str = '\n   Error during the SYRTHES preparation step\n'
            if retval == 1:
                err_str += '   Error during data copy\n'
            elif retval == 2:
                err_str += '   Error during syrthes compilation and link\n'
                # Copy source to results directory, as no run is possible
                for f in ['src', 'compile.log']:
                    self.copy_result(f)
            raise RunCaseError(err_str)

        # Set executable

        self.solver_path = os.path.join('.', 'syrthes')

        self.__ld_library_path_restore__(ld_library_path_save)

    #---------------------------------------------------------------------------

    def preprocess(self):
        """
        Partition mesh for parallel run if required by user
        """

        ld_library_path_save = os.getenv('LD_LIBRARY_PATH')
        os.environ['LD_LIBRARY_PATH'] = self.ld_library_path

        # Rebuild Syrthes case now that number of processes is known
        self.__init_syrthes_case__()

        # Sumary of the parameters
        self.syrthes_case.dump()

        # Initialize output file if needed
        self.syrthes_case.logfile_init()

        # Pre-processing (including partitioning only if SYRTHES
        # computation is done in parallel)
        retval = self.syrthes_case.preprocessing()
        if retval != 0:
            err_str = '\n  Error during the SYRTHES preprocessing step\n'
            raise RunCaseError(err_str)

        self.__ld_library_path_restore__(ld_library_path_save)

    #---------------------------------------------------------------------------

    def copy_results(self):
        """
        Retrieve results from the execution directory
        """

        ld_library_path_save = os.getenv('LD_LIBRARY_PATH')
        os.environ['LD_LIBRARY_PATH'] = self.ld_library_path

        # Post-processing
        if self.syrthes_case.post_mode != None:
          retval = self.syrthes_case.postprocessing(mode = \
                   self.syrthes_case.post_mode)
        else:
          retval = 0

        if retval != 0:
            err_str = '\n   Error during SYRTHES postprocessing\n'
            raise RunCaseError(err_str)


        if self.exec_dir == self.result_dir:
            return

        retval = self.syrthes_case.save_results(save_dir = self.result_dir,
                                                horodat = False,
                                                overwrite = True)
        if retval != 0:
            err_str = '\n   Error saving SYRTHES results\n'
            raise RunCaseError(err_str)

        self.__ld_library_path_restore__(ld_library_path_save)

    #---------------------------------------------------------------------------

    def summary_info(self, s):
        """
        Output summary data into file s
        """

        base_domain.summary_info(self, s)

        if self.solver_path:
            s.write('    SYRTHES      : ' + self.solver_path + '\n')

#-------------------------------------------------------------------------------

# Cathare coupling

class cathare_domain(domain):
    """
    Class specific to coupling with CATHARE.
    """

    #---------------------------------------------------------------------------

    def __init__(self,
                 package,                     # main package
                 package_compute = None,      # package for compute environment
                 name = None,                 # domain name
                 n_procs_weight = None,       # recommended number of processes
                 n_procs_min = None,          # min. number of processes
                 n_procs_max = None,          # max. number of processes
                 logging_args = None,         # command-line options for logging
                 param = None,                # XML parameters file
                 prefix = None,               # installation prefix
                 cathare_case_file = None,    # Cathare case .dat file
                 neptune_cfd_dom   = None,    # neptune_cfd domain
                 adaptation = None):          # HOMARD adaptation script

        domain.__init__(self,
                        package,
                        package_compute,
                        name,
                        n_procs_weight,
                        n_procs_min,
                        n_procs_max,
                        logging_args,
                        param,
                        prefix,
                        adaptation)

        # If not provided, setup is automatically set to setup.xml
        if self.param == None:
            self.param = "setup.xml"

        self.cathare_case_file = cathare_case_file
        self.neptune_cfd_dom   = neptune_cfd_dom

    #---------------------------------------------------------------------------

    def read_parameter_file(self, param):

        # Getting the .xml from NEPTUNE_CFD, since all definitions are made
        # within its GUI and paramfile

        # First we check if code is launched using the run command,
        # hence in the main folder. If using submit, then  _spath
        # corresponds to the execution folder, so we remove the DATA
        _spath = os.path.split(self.case_dir)[0]
        nept_paramfile = os.path.join(_spath,
                                      self.neptune_cfd_dom,
                                      "DATA",
                                      param)

        if not os.path.exists(nept_paramfile):
            nept_paramfile = os.path.join(_spath,
                                          self.neptune_cfd_dom,
                                          param)
            if not os.path.exists(nept_paramfile):
                raise Exception("Could not find NEPTUNE_CFD data file.")

        ofile = open(nept_paramfile,'r').readlines()
        nfile = open(os.path.join(self.data_dir, param),'w')
        for line in ofile:
            if 'api_type' not in line:
                nfile.write(line)
            else:
                nfile.write('       <api_type>2</api_type>')

        nfile.close()

        if param != None:
            version_str = '2.0'
            P = cs_xml_reader.Parser(os.path.join(self.data_dir, param),
                                     version_str = version_str)
            params = P.getParams()
            for k in list(params.keys()):
                self.__dict__[k] = params[k]

            self.param = param

            solver_dir = self.package_compute.get_dir("pkglibexecdir")
            solver_name = "nc_solver" + self.package_compute.config.exeext
            self.solver_path = os.path.join(solver_dir, solver_name)

    #---------------------------------------------------------------------------

    def compile_cathare2_lib(self):
        """
        Copy the Cathare .dat file and the different .f files to the execution
        folder. Then generate the .so lib based on the ICoCo API of
        CATHARE2.
        """

        msg = "Generating CATHARE2 .so file\n" \
              "----------------------------\n\n"
        sys.stdout.write(msg)
        sys.stdout.flush()

        import subprocess

        orig = os.getcwd()

        os.chdir(self.exec_dir)

        cathare_compilation_files = [self.cathare_case_file]

        dir_files = os.listdir(self.data_dir)
        for f in fnmatch.filter(dir_files, '*.f'):
            cathare_compilation_files.append(f)

        for f in cathare_compilation_files:
            src_file  = os.path.join(self.data_dir, f)
            dest_file = os.path.join(self.exec_dir, f)
            shutil.copy2(src_file, dest_file)

        # Generate the .so creation command (hence reducing dependencies)
        config = configparser.ConfigParser()
        config.read(self.package.get_configfiles())
        shell_cmd = 'export v25_3="%s"\n' % (config.get('install', 'cathare'))
        shell_cmd+= 'jdd_CATHARE="%s"\n' % (self.cathare_case_file)
        shell_cmd+= 'export LD_LIBRARY_PATH=${v25_3}/lib:$LD_LIBRARY_PATH\n'
        shell_cmd+= 'export LD_LIBRARY_PATH=${v25_3}/ICoCo/lib:$LD_LIBRARY_PATH\n'
        shell_cmd+= 'make -f ${v25_3}/ICoCo/Makefile_gad clean\n'
        shell_cmd+= '${v25_3}/unix-procedur/vers.unix\n'
        shell_cmd+= 'DATAFILE=${jdd_CATHARE} make -f ${v25_3}/ICoCo/Makefile_gad lib\n'

        # Shell
        user_shell = os.getenv('SHELL')
        if not user_shell:
            user_shell = '/bin/sh'

        # log
        log = open('cathare2_so_generation.log', 'w')

        p = subprocess.Popen(shell_cmd,
                             shell=True,
                             executable=user_shell,
                             stdout=log,
                             stderr=log,
                             universal_newlines=True)

        log.close()

        output, errors = p.communicate()

        if p.returncode != 0:
            self.error = 'compile cathare2 lib'
            self.error_long = 'Compilation of cathare2 .so library based on '
            self.error_long+= '%s file failed.' % (self.cathare_case_file)
            self.error_long+= ' Check "cathare2_so_generation.log' \


        os.chdir(orig)

    #---------------------------------------------------------------------------

    def compile_and_link(self):
        """
        Compile and link user subroutines if needed.
        """

        # Generate the CATHARE lib first
        self.compile_cathare2_lib()

        # Then compile NCFD source files if needed
        super(cathare_domain, self).compile_and_link()

    #---------------------------------------------------------------------------

    def solver_command(self, **kw):
        """
        Returns a tuple indicating the script's working directory,
        executable path, and associated command-line arguments.
        """

        wd, exec_path, args = super(cathare_domain, self).solver_command()
        args += " --c2-wrapper"

        return wd, exec_path, args

#-------------------------------------------------------------------------------

# Coupling with a general Python-based code

class python_domain(base_domain):

    #---------------------------------------------------------------------------

    def __init__(self,
                 package,
                 cmd_line = None,
                 name = None,
                 script_name = 'partner_script.py',
                 log_file = None,
                 n_procs_weight = None,
                 n_procs_min = 1,
                 n_procs_max = None,
                 n_threads = 1):

        base_domain.__init__(self,
                             package,
                             name,
                             n_procs_weight,
                             n_procs_min,
                             n_procs_max)

        self.cmd_line = cmd_line
        self.logfile  = log_file
        if self.logfile == None:
            self.logfile = 'python.log'

        self.data_file = None

        self.solver_path = self.package.config.python
        self.exec_solver = True
        self.nthreads = n_threads

        self.script_name = script_name

    #---------------------------------------------------------------------------

    def set_case_dir(self, case_dir, staging_dir = None):

        base_domain.set_case_dir(self, case_dir, staging_dir)

        self.data_dir = os.path.join(self.case_dir, "DATA")
        self.src_dir  = None
        self.result_dir = os.path.join(self.case_dir, "RESU")

    #---------------------------------------------------------------------------

    def set_exec_dir(self, exec_dir):

        if os.path.isabs(exec_dir):
            self.exec_dir = exec_dir
        else:
            self.exec_dir = os.path.join(self.case_dir, 'RESU', exec_dir)

        self.exec_dir = os.path.join(self.exec_dir, self.name)

        if not os.path.isdir(self.exec_dir):
            os.mkdir(self.exec_dir)

    #---------------------------------------------------------------------------

    def set_result_dir(self, name, given_dir = None):

        if given_dir == None:
            self.result_dir = os.path.join(self.result_dir,
                                           'RESU_' + self.name,
                                           name)
        else:
            self.result_dir = os.path.join(given_dir, self.name)

        if not os.path.isdir(self.result_dir):
            os.makedirs(self.result_dir)

    #---------------------------------------------------------------------------

    def copy_data(self):
        """
        Copy data to run directory
        """

        dir_files = os.listdir(self.data_dir)

        for f in dir_files:
            src = os.path.join(self.data_dir, f)
            if os.path.isfile(src):
                shutil.copy2(src, os.path.join(self.exec_dir, f))

    #---------------------------------------------------------------------------

    def preprocess(self):
        """
        Preprocess dummy function: Does nothing for a standard python script
        """

        # Nothing to do

    #---------------------------------------------------------------------------

    def copy_results(self):
        """
        Copy results dummy function: Does nothing for a standard python script
        """
        # Nothing to do
        return

    #---------------------------------------------------------------------------

    def summary_info(self, s):
        """
        output summary data into file s
        """

        base_domain.summary_info(self, s)

    #---------------------------------------------------------------------------

    def solver_command(self, **kw):
        """
        Returns a tuple indicating the script's working directory,
        executable path, and associated command-line arguments.
        """

        wd = enquote_arg(self.exec_dir)              # Working directory
        # Executable
        exec_path = enquote_arg(self.solver_path) \
                  + ' ' + self.script_name

        # Build kernel command-line arguments

        args = ''

        if self.data_file:
            args += ' -d ' + enquote_arg(self.data_file)

        args += ' -n ' + str(self.n_procs)

        args += ' --name ' + enquote_arg(self.name)

        # Output to a logfile
        args += ' --log ' + enquote_arg(self.logfile)

        return wd, exec_path, args

#-------------------------------------------------------------------------------
# End
#-------------------------------------------------------------------------------



# -*- coding: utf-8 -*-

#-------------------------------------------------------------------------------

# This file is part of Code_Saturne, a general-purpose CFD tool.
#
# Copyright (C) 1998-2021 EDF S.A.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.

#-------------------------------------------------------------------------------

"""
This module defines the gas combustion thermal flow modelling management.

This module contains the following classes and function:
- GasCombustionModel
- GasCombustionTestCase
- ThermochemistryData
"""

#-------------------------------------------------------------------------------
# Library modules import
#-------------------------------------------------------------------------------

import sys, unittest

#-------------------------------------------------------------------------------
# Application modules import
#-------------------------------------------------------------------------------

from code_saturne.model.Common import *
from code_saturne.model.XMLvariables import Variables, Model
from code_saturne.model.ThermalScalarModel import ThermalScalarModel
from code_saturne.model.TurbulenceModel import TurbulenceModel
from code_saturne.model.ThermalRadiationModel import ThermalRadiationModel
from code_saturne.model.FluidCharacteristicsModel import FluidCharacteristicsModel
from code_saturne.model.NumericalParamEquationModel import NumericalParamEquationModel
from code_saturne.model.LocalizationModel import LocalizationModel
from code_saturne.model.Boundary import Boundary

#-------------------------------------------------------------------------------
# Gas combustion model class
#-------------------------------------------------------------------------------

class GasCombustionModel(Variables, Model):
    """
    """
    def __init__(self, case):
        """
        Constructor.
        """
        self.case = case

        nModels          = self.case.xmlGetNode('thermophysical_models')
        self.node_turb   = nModels.xmlInitNode('turbulence',        'model')
        self.node_gas    = nModels.xmlInitNode('gas_combustion',    'model')
        self.node_coal   = nModels.xmlInitNode('solid_fuels',       'model')
        self.node_joule  = nModels.xmlInitNode('joule_effect',      'model')
        self.node_atmo   = nModels.xmlInitNode('atmospheric_flows', 'model')
        self.node_prop   = self.case.xmlGetNode('physical_properties')
        self.node_fluid  = self.node_prop.xmlInitNode('fluid_properties')

        self.gasCombustionModel = ('off', 'ebu', 'd3p','lwp')
        self.d3p_list = ("adiabatic", "extended")
        self.ebu_list = ("spalding", "enthalpy_st", "mixture_st", "enthalpy_mixture_st")
        self.lwp_list = ("2-peak_adiabatic", "2-peak_enthalpy",
                         "3-peak_adiabatic", "3-peak_enthalpy",
                         "4-peak_adiabatic", "4-peak_enthalpy")

        self.sootModels = ('off', 'constant_soot_yield', 'moss')


    def defaultGasCombustionValues(self):
        """
        Return in a dictionnary which contains default values.
        """
        default = {}
        default['model'] = "off"

        model = self.getGasCombustionModel()
        if model == 'd3p':
            default['option'] = "adiabatic"
        elif model == 'ebu':
            default['option'] = "spalding"
        elif model == 'lwp':
            default['option'] = "2-peak_adiabatic"
        elif model == 'off':
            default['option'] = "off"

        return default


    @Variables.noUndo
    def getAllGasCombustionModels(self):
        """
        Return all defined gas combustion models in a tuple.
        """
        return self.gasCombustionModel


    def gasCombustionModelsList(self):
        """
        Create a tuple with the available gas combustion models.
        """
        return self.gasCombustionModel


    @Variables.undoGlobal
    def setGasCombustionModel(self, model):
        """
        Update the gas combustion model markup from the XML document.
        """
        self.isInList(model, self.gasCombustionModelsList())
        node_prop   = self.case.xmlGetNode('physical_properties')
        node_fluid  = node_prop.xmlInitNode('fluid_properties')

        old_model = self.node_gas['model']
        ThermalScalarModel(self.case).setThermalModel('off')

        if model == 'off':
            self.node_gas['model'] = model
            self.node_gas['option'] = "off"
            ThermalRadiationModel(self.case).setRadiativeModel('off')
            for tag in ('variable',
                        'property',
                        'reference_mass_molar',
                        'reference_temperature',
                        'soot_model'):
                for node in self.node_gas.xmlGetNodeList(tag):
                    node.xmlRemoveNode()
            for zone in LocalizationModel('BoundaryZone', self.case).getZones():
                if zone.getNature() == "inlet":
                    Boundary("inlet", zone.getLabel(), self.case).deleteGas()

            node_fluid.xmlRemoveChild('property', name='dynamic_diffusion')

        else:
            self.node_gas['model'] = model
            self.node_coal['model']  = 'off'
            self.node_joule['model'] = 'off'
            self.setNewFluidProperty(node_fluid, 'dynamic_diffusion')

            if old_model != model:
                for zone in LocalizationModel('BoundaryZone', self.case).getZones():
                    if zone.getNature() == "inlet":
                        Boundary("inlet", zone.getLabel(), self.case).deleteGas()

        if model != 'd3p':
            self.node_fluid.xmlRemoveChild('reference_oxydant_temperature')
            self.node_fluid.xmlRemoveChild('reference_fuel_temperature')

        self.createModel()


    @Variables.noUndo
    def getGasCombustionModel(self):
        """
        Return the current gas combustion model.
        """
        model = self.node_gas['model']
        if model not in self.gasCombustionModelsList():
            model = 'off'
            self.setGasCombustionModel(model)

        return model


    @Variables.noUndo
    def getGasCombustionOption(self):
        """
        Return the current gas combustion option.
        """
        option = self.node_gas['option']
        if option == None:
            option = self.defaultGasCombustionValues()['option']
            self.setGasCombustionOption(option)

        model = self.getGasCombustionModel()
        if model == 'd3p':
            if option not in self.d3p_list:
                option = self.defaultGasCombustionValues()['option']
                self.setGasCombustionOption(option)
        elif model == 'ebu':
            if option not in self.ebu_list:
                option = self.defaultGasCombustionValues()['option']
                self.setGasCombustionOption(option)
        elif model == 'lwp':
            if option not in self.lwp_list:
                option = self.defaultGasCombustionValues()['option']
                self.setGasCombustionOption(option)
        elif model == 'off':
            option = 'off'
        return option


    @Variables.undoGlobal
    def setGasCombustionOption(self, option):
        """
        Return the current gas combustion option.
        """
        model = self.getGasCombustionModel()
        if model == 'd3p':
            self.isInList(option, self.d3p_list)
        elif model == 'ebu':
            self.isInList(option, self.ebu_list)
        elif model == 'lwp':
            self.isInList(option, self.lwp_list)
        elif model == 'off':
            self.isInList(option, ('off'))
        self.node_gas['option'] = option
        option = self.node_gas['option']
        self.createModel()


    def __createModelScalarsList(self , model):
        """
        Private method
        Create model scalar list
        """
        option = self.getGasCombustionOption()
        list_options = ["3-peak_adiabatic", "3-peak_enthalpy",
                        "4-peak_adiabatic", "4-peak_enthalpy"]
        acceptable_options = ["2-peak_enthalpy", "3-peak_enthalpy",
                              "4-peak_enthalpy"]
        lst = []

        ThermalScalarModel(self.case).setThermalModel('off')

        if model == 'd3p':
            lst.append("mixture_fraction")
            lst.append("mixture_fraction_variance")
            if option == 'extended':
                ThermalScalarModel(self.case).setThermalModel('enthalpy')
        elif model == 'ebu':
            lst.append("fresh_gas_fraction")
            if option == "mixture_st" or option =="enthalpy_mixture_st":
                lst.append("mixture_fraction")
            elif option == "enthalpy_st" or option =="enthalpy_mixture_st":
                ThermalScalarModel(self.case).setThermalModel('enthalpy')
        elif model == 'lwp':
            lst.append("mixture_fraction")
            lst.append("mixture_fraction_variance")
            lst.append("mass_fraction")
            lst.append("mass_fraction_covariance")
            if option in list_options:
                lst.append("mass_fraction_variance")
            if option in acceptable_options:
                ThermalScalarModel(self.case).setThermalModel('enthalpy')
        return lst


    def __createModelPropertiesList(self, model):
        """
        Private method
        Create model properties
        """
        lst = []
        lst.append("temperature")
        lst.append("ym_fuel")
        lst.append("ym_oxyd")
        lst.append("ym_prod")
        if model == 'lwp':
            lst.append("source_term")
            lst.append("molar_mass")
            ndirac = self.getNdirac()
            for idirac in range(ndirac):
                lst.append("rho_local_" + str(idirac + 1))
                lst.append("temperature_local_" + str(idirac + 1))
                lst.append("ym_local_" + str(idirac + 1))
                lst.append("w_local_" + str(idirac + 1))
                lst.append("amplitude_local_" + str(idirac + 1))
                lst.append("chemical_st_local_" + str(idirac + 1))
                lst.append("molar_mass_local_" + str(idirac + 1))
        return lst


    def __createModelScalars(self , model):
        """
        Private method
        Create model scalar
        """
        previous_list = []
        nodes = self.node_gas.xmlGetChildNodeList('variable')
        for node in nodes:
            previous_list.append(node['name'])

        if model == "off":
            for node in nodes:
                node.xmlRemoveNode()
        else:
            new_list = self.__createModelScalarsList(model)
            for name in previous_list:
                if name not in new_list:
                    self.node_gas.xmlRemoveChild('variable',  name = name)

            for name in new_list:
                if name not in previous_list:
                    self.setNewVariable(self.node_gas, name, tpe='var_model', label=name)

            NPE = NumericalParamEquationModel(self.case)
            for node in self.node_gas.xmlGetChildNodeList('variable'):
                name = node['name']
                NPE.setBlendingFactor(name, 0.)
                NPE.setScheme(name, 'upwind')
                NPE.setFluxReconstruction(name, 'off')

                if self.getGasCombustionModel() == "d3p":
                    if name == "mixture_fraction":
                        NPE.setMinValue(name, 0.)
                        NPE.setMaxValue(name, 1.)
                    elif name == "mixture_fraction_variance":
                        NPE.setMinValue(name, 0.)
                        NPE.setMaxValue(name, 1.e+12)
                        node.xmlSetData('variance', "mixture_fraction")


    @Variables.noUndo
    def getNdirac(self):
        """
        """
        option = self.getGasCombustionOption()
        self.isInList(option, self.lwp_list)
        if option == '2-peak_adiabatic' or option == '2-peak_enthalpy':
            ndirac = 2
        if option == '3-peak_adiabatic' or option == '3-peak_enthalpy':
            ndirac = 3
        if option == '4-peak_adiabatic' or option == '4-peak_enthalpy':
            ndirac = 4
        return ndirac


    def __createModelProperties(self, model):
        """
        Private method
        Create model properties
        """
        previous_list = []
        nodes = self.node_gas.xmlGetChildNodeList('property')
        if model == "off":
            for node in nodes:
                node.xmlRemoveNode()
        else:
            for node in nodes:
                previous_list.append(node['name'])

            new_list = self.__createModelPropertiesList(model)
            for name in previous_list:
                if name not in new_list:
                    self.node_gas.xmlRemoveChild('property',  name = name)

            for name in new_list:
                if name not in previous_list:
                    self.setNewProperty(self.node_gas, name)


    def createModel (self) :
        """
        Private method
        Create scalars and properties when gas combustion is selected
        """
        model = self.getGasCombustionModel()
        self.__createModelScalars(model)
        self.__createModelProperties(model)


    @Variables.noUndo
    def getThermoChemistryDataFileName(self):
        """
        Get name for properties data (return None if not defined)i
        """
        f = self.node_gas.xmlGetString('data_file')
        return f


    @Variables.undoLocal
    def setThermoChemistryDataFileName(self, name):
        """
        Put name for properties data and load file for number gaz and radiative model
        """
        self.node_gas.xmlSetData('data_file', name)

    def _defaultValues(self):
        """
        default values
        """
        self.default = {}
        self.default['thermodynamical_pressure'] = 'off'
        self.default['soot_model']               = 'off'
        self.default['soot_density']             = 0.0
        self.default['soot_fraction']            = 0.0
        return self.default

    @Variables.noUndo
    def getUniformVariableThermodynamicalPressure(self):
        """
        Return status of uniform variable thermodynamical pressure
        """
        node = self.node_gas.xmlInitNode('thermodynamical_pressure', 'status')
        status = node['status']
        if not status:
            status = self._defaultValues()['thermodynamical_pressure']
            self.setUniformVariableThermodynamicalPressure(status)
        return status

    @Variables.undoLocal
    def setUniformVariableThermodynamicalPressure(self, status):
        """
        Put status of uniform variable thermodynamical pressure
        """
        self.isOnOff(status)
        node = self.node_gas.xmlInitNode('thermodynamical_pressure', 'status')
        node['status'] = status

    @Variables.noUndo
    def getSootModel(self):
        """
        Return value of attribute model
        """
        node = self.node_gas.xmlInitChildNode('soot_model', 'model')
        model = node['model']
        if model not in self.sootModels:
            model = self._defaultValues()['soot_model']
            self.setSootModel(model)
        return model

    @Variables.undoGlobal
    def setSootModel(self, model):
        """
        Put value of attribute model to soot model
        """
        self.isInList(model, self.sootModels)
        node  = self.node_gas.xmlInitChildNode('soot_model', 'model')
        node['model'] = model
        if model == 'moss':
            self.node_gas.xmlRemoveChild('soot_fraction')
        if model == 'off':
            self.node_gas.xmlRemoveChild('soot_density')
            self.node_gas.xmlRemoveChild('soot_fraction')

    @Variables.noUndo
    def getSootDensity(self):
        """
        Return value of soot density
        """
        val = self.node_gas.xmlGetDouble('soot_density')
        if val == None:
            val = self._defaultValues()['soot_density']
            self.setSootDensity(val)
        return val

    @Variables.undoGlobal

    @Variables.undoGlobal
    def setSootDensity(self, val):
        """
        Put value of soot density
        """
        self.isPositiveFloat(val)
        self.node_soot = self.node_gas.xmlGetNode('soot_model')
        self.node_soot.xmlSetData('soot_density', val)

    @Variables.noUndo
    def getSootFraction(self):
        """
        Return value of soot fraction
        """
        val = self.node_gas.xmlGetDouble('soot_fraction')
        if val == None:
            val = self._defaultValues()['soot_fraction']
            self.setSootFraction(val)
        return val

    @Variables.undoGlobal
    def setSootFraction(self, val):
        """
        Put value of soot fraction
        """
        self.isPositiveFloat(val)
        self.node_soot = self.node_gas.xmlGetNode('soot_model')
        self.node_soot.xmlSetData('soot_fraction', val)

#-------------------------------------------------------------------------------
# Gas combustion test case
#-------------------------------------------------------------------------------


class GasCombustionTestCase(unittest.TestCase):
    """
    """
    def setUp(self):
        """This method is executed before all "check" methods."""
        from code_saturne.model.XMLengine import Case, XMLDocument
        from code_saturne.model.XMLinitialize import XMLinit
        GuiParam.lang = 'en'
        self.case = Case(None)
        XMLinit(self.case).initialize()
        self.doc = XMLDocument()

    def tearDown(self):
        """This method is executed after all "check" methods."""
        del self.case
        del self.doc

    def xmlNodeFromString(self, string):
        """Private method to return a xml node from string"""
        return self.doc.parseString(string).root()

    def checkGasCombustionInstantiation(self):
        """
        Check whether the gasCombustionModel class could be instantiated
        """
        model = None
        model = GasCombustionModel(self.case)
        assert model != None, 'Could not instantiate GasCombustionModel'


def suite():
    testSuite = unittest.makeSuite(GasCombustionTestCase, "check")
    return testSuite


def runTest():
    print("GasCombustionTestCase - TODO**************")
    runner = unittest.TextTestRunner()
    runner.run(suite())


#-------------------------------------------------------------------------------
# Thermochemistry data class
#-------------------------------------------------------------------------------

class ThermochemistryData(Model):
    """
    Useful methods to create a Janaf File with the GUI
    """
    def __init__(self, case):

        """
        Constuctor.
        """
        self.case = case

        nModels              = self.case.xmlGetNode('thermophysical_models')
        self.node_thermodata = nModels.xmlInitNode('Thermochemistry_data')

        # Parameters to create the Thermochemistry Data File
        # Available Chemical Elements (WARNING must be in upper case)
        self.ChemicalElem = ['C', 'H', 'O', 'N', 'CL']
        # Molar mass for the Chemical Elements
        # (WARNING must be in upper case)
        self.MolarMass = {}
        self.MolarMass['C']  = 0.012
        self.MolarMass['H']  = 0.001
        self.MolarMass['O']  = 0.016
        self.MolarMass['N']  = 0.014
        self.MolarMass['CL'] = 0.0354
        # Number of known global species, (ngazg in colecd.f90)
        # = 2 -> Automatic equilibrium of the reaction
        # = 3 -> Equilibrium defined by the user
        # NumberOfKnownGlobalSpecies is updated in WriteThermochemistryDataFile
        self.NumberOfKnownGlobalSpecies = 3
        # Always 1 reaction (see colecd.f90)
        self.NbReaction = 1
        self.Error_GUI = False


    def defaultParamforTabu(self):
        """
        Return in a dictionnary which contains default values 
        for parameters necessarry to create the Janaf File
        """
        default = {}
        default['CreateThermoDataFile'] = 'off'
        default['user_mode_chem'] = 'auto'
        default['NbPointsTabu'] = 10
        default['MaximumTemp'] = 3000.0
        default['MinimumTemp'] = 273.0

        return default


    @Variables.noUndo
    def getCreateThermoDataFile(self):
        """
        Return status of CreateThermoDataFile
        """
        node = self.node_thermodata.xmlInitNode('CreateThermoDataFile', 'status')
        status = node['status']
        if not status:
            status = self.defaultParamforTabu()['CreateThermoDataFile']
            self.setCreateThermoDataFile(status)
        return status


    @Variables.undoLocal
    def setCreateThermoDataFile(self, status):
        """
        Put status of CreateThermoDataFile
        """
        self.isOnOff(status)
        node = self.node_thermodata.xmlInitNode('CreateThermoDataFile', 'status')
        node['status'] = status
        if status != 'on':
            self.node_thermodata.xmlRemoveChild('NbPointsTabu')
            self.node_thermodata.xmlRemoveChild('MaximumTemp')
            self.node_thermodata.xmlRemoveChild('MinimumTemp')
            self.node_thermodata.xmlRemoveChild('user_mode_chem')


    @Variables.noUndo
    def getUserModeForChemicalReaction(self):
        """
        Return the mode to build the Janaf File
        """
        node = self.node_thermodata.xmlInitChildNode('user_mode_chem', 'choice')
        model = node['choice']
        if not model:
            model = self.defaultParamforTabu()['user_mode_chem']
            self.setUserModeForChemicalReaction(model)
        return model


    @Variables.undoGlobal
    def setUserModeForChemicalReaction(self, model):
        """
        Put the mode to build the Janaf File
        """
        self.node_mode_chem = self.node_thermodata.xmlInitChildNode('user_mode_chem', 'choice')
        self.node_mode_chem['choice'] = model

        if model == 'user':
            self.node_mode_chem.xmlRemoveChild('Fuel')
            self.node_mode_chem.xmlRemoveChild('VolPropO2')
            self.node_mode_chem.xmlRemoveChild('VolPropN2')
            self.node_mode_chem.xmlRemoveChild('COyield')
            self.node_mode_chem.xmlRemoveChild('CSyield')
        if model == 'auto':
            nodes = self.node_mode_chem.xmlGetChildNodeList('variable')
            for node in nodes:
                node.xmlRemoveNode()


    @Variables.noUndo
    def getNbPointsTabu(self):
        """
        Return value of NbPointsTabu
        """
        NbPointsTabu = self.node_thermodata.xmlGetInt('NbPointsTabu')
        if NbPointsTabu == None:
            NbPointsTabu = self.defaultParamforTabu()['NbPointsTabu']
            self.setNbPointsTabu(NbPointsTabu)

        return NbPointsTabu


    @Variables.undoGlobal
    def setNbPointsTabu(self, value):
        """
        Put value of NbPointsTabu
        """
        self.isInt(value)
        self.node_thermodata.xmlSetData('NbPointsTabu', value)


    @Variables.noUndo
    def getMaximumTemp(self):
        """
        Return value of MaximumTemp
        """
        MaximumTemp = self.node_thermodata.xmlGetDouble('MaximumTemp')
        if MaximumTemp == None:
            MaximumTemp = self.defaultParamforTabu()['MaximumTemp']
            self.setMaximumTemp(MaximumTemp)

        return MaximumTemp


    @Variables.undoGlobal
    def setMaximumTemp(self, value):
        """
        Put value of MaximumTemp
        """
        self.isFloat(value)
        self.node_thermodata.xmlSetData('MaximumTemp', value)


    @Variables.noUndo
    def getMinimumTemp(self):
        """
        Return value of MinimumTemp
        """
        MinimumTemp = self.node_thermodata.xmlGetDouble('MinimumTemp')
        if MinimumTemp == None:
            MinimumTemp = self.defaultParamforTabu()['MinimumTemp']
            self.setMinimumTemp(MinimumTemp)

        return MinimumTemp


    @Variables.undoGlobal
    def setMinimumTemp(self, value):
        """
        Put value of MinimumTemp
        """
        self.isFloat(value)
        self.node_thermodata.xmlSetData('MinimumTemp', value)


    def defaultSpeciesProperties(self):
        """
        Return the default properties for a species (mode "auto" and "user")
        """
        default = {}
        default['species_label']       = "species"
        default['chemical_formula']    = "CHON"
        default['fuel_composition']    = 0.0
        default['oxi_composition']     = 0.0
        default['prod_composition']    = 0.0
        default['coeff_absorption']    = 0.35
        default['fuel']                = "C4H10"
        default['volPropO2']           = 0.21
        default['volPropN2']           = 0.79
        default['COyield']             = 0.0
        default['CSyield']             = 0.0

        return default


    @Variables.noUndo
    def getChemicalFormulaFuel(self):
        """
        Return the chemical formula for the Fuel
        """
        ChemicalFormula = self.node_thermodata.xmlGetString('Fuel')
        if ChemicalFormula == "":
            ChemicalFormula = self.defaultSpeciesProperties()['fuel']
            self.setChemicalFormulaFuel(ChemicalFormula)

        return ChemicalFormula


    @Variables.undoGlobal
    def setChemicalFormulaFuel(self, chemical_formula):
        """
        Put the chemical formula for the Fuel
        """
        self.node_mode_chem = self.node_thermodata.xmlGetNode('user_mode_chem')
        self.node_mode_chem.xmlSetData('Fuel', chemical_formula)


    @Variables.noUndo
    def getVolPropO2(self):
        """
        Return the value of VolPropO2
        """
        VolPropO2 = self.node_thermodata.xmlGetDouble('VolPropO2')
        if VolPropO2 == None:
            VolPropO2 = self.defaultSpeciesProperties()['volPropO2']
            self.setVolPropO2(VolPropO2)

        return VolPropO2


    @Variables.undoGlobal
    def setVolPropO2(self, value):
        """
        Put the value of VolPropO2
        """
        self.isFloat(value)
        self.node_mode_chem = self.node_thermodata.xmlGetNode('user_mode_chem')
        self.node_mode_chem.xmlSetData('VolPropO2', value)


    @Variables.noUndo
    def getVolPropN2(self):
        """
        Return the value of VolPropN2
        """
        VolPropN2 = self.node_thermodata.xmlGetDouble('VolPropN2')
        if VolPropN2 == None:
            VolPropN2 = self.defaultSpeciesProperties()['volPropN2']
            self.setVolPropN2(VolPropN2)

        return VolPropN2


    @Variables.undoGlobal
    def setVolPropN2(self, value):
        """
        Put the value of VolPropN2
        """
        self.isFloat(value)
        self.node_mode_chem = self.node_thermodata.xmlGetNode('user_mode_chem')
        self.node_mode_chem.xmlSetData('VolPropN2', value)


    @Variables.noUndo
    def getCOyield(self):
        """
        Return the value of COyield
        """
        COyield = self.node_thermodata.xmlGetDouble('COyield')
        if COyield == None:
            COyield = self.defaultSpeciesProperties()['COyield']
            self.setCOyield(COyield)

        return COyield


    @Variables.undoGlobal
    def setCOyield(self, value):
        """
        Put the value of COyield
        """
        self.isFloat(value)
        self.node_mode_chem = self.node_thermodata.xmlGetNode('user_mode_chem')
        self.node_mode_chem.xmlSetData('COyield', value)


    @Variables.noUndo
    def getCSyield(self):
        """
        Return the value of CSyield
        """
        CSyield = self.node_thermodata.xmlGetDouble('CSyield')
        if CSyield == None:
            CSyield = self.defaultSpeciesProperties()['CSyield']
            self.setCSyield(CSyield)

        return CSyield


    @Variables.undoGlobal
    def setCSyield(self, value):
        """
        Put the value of CSyield
        """
        self.isFloat(value)
        self.node_mode_chem = self.node_thermodata.xmlGetNode('user_mode_chem')
        self.node_mode_chem.xmlSetData('CSyield', value)


    def __defaultSpeciesLabel(self, species_name=None):
        """
        Private method.
        Return a default label for a new species.
        """
        __coef = {}
        for l in self.getSpeciesNamesList():
            __coef[l] = l
        length = len(__coef)
        Lspe = self.defaultSpeciesProperties()['species_label']

        # new species: default value 

        if not species_name:
            if length != 0:
                i = 1
                while (Lspe + str(i)) in list(__coef.values()):
                    i = i + 1
                num = str(i)
            else:
                num = str(1)
            species_name = Lspe + num
        return species_name


    @Variables.undoGlobal
    def addSpecies(self, name=None):
        """
        Public method.
        Input a new species I{name}
        """

        c = self.__defaultSpeciesLabel(name)

        self.node_mode_chem = self.node_thermodata.xmlGetNode('user_mode_chem')

        if c not in self.getSpeciesNamesList():
            self.node_mode_chem.xmlInitNode('variable', label=c, name=c)
            self.setSpeciesChemicalFormula(c, self.defaultSpeciesProperties()['chemical_formula'])
            self.setCompFuel(c, self.defaultSpeciesProperties()['fuel_composition'])
            self.setCompOxi(c, self.defaultSpeciesProperties()['oxi_composition'])
            self.setCompProd(c, self.defaultSpeciesProperties()['prod_composition'])
            self.setCoeffAbsorp(c, self.defaultSpeciesProperties()['coeff_absorption'])

        return c


    @Variables.undoGlobal
    def updateSpecies(self, data):
        """
        Public method.
        Update the XML from the tableView
        """
        c = data[0]

        self.node_mode_chem = self.node_thermodata.xmlGetNode('user_mode_chem')

        self.node_mode_chem.xmlInitNode('variable', label=c, name=c)
        self.setSpeciesChemicalFormula(c, data[1])
        self.setCompFuel(c, data[2])
        self.setCompOxi(c, data[3])
        self.setCompProd(c, data[4])
        self.setCoeffAbsorp(c, data[5])


    @Variables.undoGlobal
    def deleteSpecies(self, species_label):
        """
        Public method.
        Delete species I{name}
        """
        self.isInList(species_label, self.getSpeciesNamesList())
        node = self.node_thermodata.xmlGetNode('variable', label=species_label)
        node.xmlRemoveNode()


    @Variables.noUndo
    def getSpeciesNamesList(self):
        """
        Public method.
        Return the Species name list
        """
        self.node_mode_chem = self.node_thermodata.xmlGetNode('user_mode_chem')
        lst = []
        for node in self.node_mode_chem.xmlGetChildNodeList('variable'):
                lst.append(node['label'])
        return lst


    @Variables.noUndo
    def getSpeciesChemicalFormula(self, l):
        """
        Public method.
        Return the Chemical Formula
        """
        n = self.case.xmlGetNode('variable', label=l)
        name = n.xmlGetString('chemical_formula')
        return name


    @Variables.undoGlobal
    def setSpeciesChemicalFormula(self, species_label, species_name):
        """
        Put the Chemical Formula
        """
        n = self.case.xmlGetNode('variable', label=species_label)
        n.xmlSetData('chemical_formula', species_name)


    @Variables.undoGlobal
    def getCompFuel(self, l):
        """
        Get Fuel Composition
        """
        n = self.case.xmlGetNode('variable', label=l)
        val = n.xmlGetDouble('fuel_composition')
        if val == None:
            val = self.defaultSpeciesProperties()['fuel_composition']
            self.setCompFuel(l, val)

        return val


    @Variables.undoGlobal
    def setCompFuel(self, species_label, CompFuel):
        """
        Put Fuel Composition
        """
        n = self.case.xmlGetNode('variable', label=species_label)
        n.xmlSetData('fuel_composition', CompFuel)


    @Variables.undoGlobal
    def getCompOxi(self, l):
        """
        Get Oxi Composition
        """
        n = self.case.xmlGetNode('variable', label=l)
        val = n.xmlGetDouble('oxi_composition')
        if val == None:
            val = self.defaultSpeciesProperties()['oxi_composition']
            self.setCompOxi(l, val)

        return val


    @Variables.undoGlobal
    def setCompOxi(self, species_label, CompOxi):
        """
        Put Oxi Composition
        """
        n = self.case.xmlGetNode('variable', label=species_label)
        n.xmlSetData('oxi_composition', CompOxi)


    @Variables.undoGlobal
    def getCompProd(self, l):
        """
        Get Product Composition
        """
        n = self.case.xmlGetNode('variable', label=l)
        val = n.xmlGetDouble('prod_composition')
        if val == None:
            val = self.defaultSpeciesProperties()['prod_composition']
            self.setCompProd(l, val)

        return val


    @Variables.undoGlobal
    def setCompProd(self, species_label, CompProd):
        """
        Put Product Composition
        """
        n = self.case.xmlGetNode('variable', label=species_label)
        n.xmlSetData('prod_composition', CompProd)


    @Variables.undoGlobal
    def getCoeffAbsorp(self, l):
        """
        Get absorption coefficient
        """
        n = self.case.xmlGetNode('variable', label=l)
        val = n.xmlGetDouble('coeff_absorption')
        if val == None:
            val = self.defaultSpeciesProperties()['coeff_absorption']
            self.setCoeffAbsorp(l, val)

        return val


    @Variables.undoGlobal
    def setCoeffAbsorp(self, species_label, CoeffAbsorp):
        """
        Put absorption coefficient
        """
        n = self.case.xmlGetNode('variable', label=species_label)
        n.xmlSetData('coeff_absorption', CoeffAbsorp)


    @Variables.undoGlobal
    def getNumberOfChemicalElem(self, ChemicalFormula):
        """
        Get the number of Chemical Element
        """

        ChemicalFormula = str(ChemicalFormula.upper())
        
        NumberOfChemElem = {}
        for Elem in self.ChemicalElem:
            NumberOfChemElem[Elem] = "0"
        
        ChemicalElem_sorted = sorted(self.ChemicalElem, key=len, reverse=True)

        for Elem in ChemicalElem_sorted :
            ReadFirstElem = True
            if Elem in ChemicalFormula :
                NumberOfChemElem[Elem] = "1"
                ReadFirstElem = True
                position = ChemicalFormula.find(Elem)+len(Elem)
                ChemicalFormula = ChemicalFormula.replace(str(Elem), ' ')
                while ChemicalFormula[position:position+1].isdigit():
                    if ReadFirstElem: 
                        NumberOfChemElem[Elem] = ChemicalFormula[position:position+1]
                    else :
                        NumberOfChemElem[Elem] = NumberOfChemElem[Elem]+ChemicalFormula[position:position+1]

                    temp = list(ChemicalFormula)
                    temp[position] = ' '
                    ChemicalFormula = "".join(temp)

                    ReadFirstElem = False
                    position=position+1
                
        return NumberOfChemElem

    @Variables.undoGlobal
    def WriteThermochemistryDataFile(self, file_path):
        """
        Write the thermochemistry Data File
        """
        # Assumption : only one reaction (in colecd : ir = 1) (parameter NbReaction)
        # in colecd : ngazg can equal to 2 or 3 (parameter NumberOfKnownGlobalSpecies)
        # if ngazg = 3 we add the parameters:
        # in colecd : igfuel and igoxy (parameters position_fuel and position_oxi) and in colecd : stoeg (parameter global_stoeg)
        # see colecd.f90 for more information

        #Comment to add at the end of each lines FR
        InfoLine = {}
        InfoLine['NumberOfSpecies'] = "Nb especes courantes"
        InfoLine['NbPointsTabu'] = "Nb de points de tabulation ENTH-TEMP"
        InfoLine['MinimumTemp'] = "TMIN"
        InfoLine['MaximumTemp'] = "TMAX"
        InfoLine['LineInfo-GaseousSpecies'] = "Especes Gazeuses"
        InfoLine['CoeffAbsorp'] = "Coeff absorption (ray)"
        InfoLine['GlobalElemCompo'] = "Nb especes elementaires"
        InfoLine['LineInfo-ChemElem'] = "Composition CHON"
        InfoLine['NumberOfKnownGlobalSpecies'] = "Nb d'especes globales connues (ici : / Fuel / Oxydant / Produits)"
        InfoLine['FuelComposition'] = "Numero espece reactive / Composition Fuel     en especes elementaires"
        InfoLine['OxiComposition'] = "Numero espece reactive / Composition Oxydant  en especes elementaires"
        InfoLine['ProdComposition'] = "Numero espece reactive / Composition Produits en especes elementaires"
        InfoLine['NbReaction'] = "Nb de reactions mises en jeu pour les especes globales"
        InfoLine['Stoeg'] = "Stoechiometrie en nb de mole especes globales"
        #----------------------------------------
        ErrorLine = {}
        ErrorLine['Error'] = "Error in the GUI\n"
        ErrorLine['No_species'] = "There are no species defined\n"
        ErrorLine['FLAG_No_Oxi'] = "There is no Oxidiser specified\n"
        ErrorLine['FLAG_No_Fuel'] = "There is no Fuel specified\n"
        ErrorLine['FLAG_No_Prod'] = "There is no Product specified\n"
        ErrorLine['FLAG_No_compo'] = " has no Nb of moles specified\n"

        #Prepare the data depending from the user mode (automatic or defined by user)
        ListOfSpecies       = []
        CompFuelDict        = {}
        CompOxiDict         = {}
        CompProdDict        = {}
        CoeffAbsorpDict     = {}
        ChemicalFormulaDict = {}
        Option_UserMode = self.getUserModeForChemicalReaction()
        if Option_UserMode == 'user':
            self.NumberOfKnownGlobalSpecies = 3
            ListOfSpecies       = self.getSpeciesNamesList()
            for label in ListOfSpecies:
                CompFuelDict[label]        = self.getCompFuel(label)
                CompOxiDict[label]         = self.getCompOxi(label)
                CompProdDict[label]        = self.getCompProd(label)
                CoeffAbsorpDict[label]     = self.getCoeffAbsorp(label)
                ChemicalFormulaDict[label] = self.getSpeciesChemicalFormula(label)
        elif Option_UserMode == 'auto':
            self.NumberOfKnownGlobalSpecies = 2
            #Fuel
            ListOfSpecies.append('fuel')
            CompFuelDict['fuel']        = 1.0
            CompOxiDict['fuel']         = 0.0
            CompProdDict['fuel']        = 0.0
            CoeffAbsorpDict['fuel']     = self.defaultSpeciesProperties()['coeff_absorption']
            ChemicalFormulaDict['fuel'] = self.getChemicalFormulaFuel()
            #O2
            ListOfSpecies.append('O2')
            CompFuelDict['O2']          = 0.0
            CompOxiDict['O2']           = 1.0
            CompProdDict['O2']          = 0.0
            CoeffAbsorpDict['O2']       = self.defaultSpeciesProperties()['coeff_absorption']
            ChemicalFormulaDict['O2']   = 'O2'
            #N2
            if self.getVolPropN2() != 0.0:
                ListOfSpecies.append('N2')
                CompFuelDict['N2']          = 0.0
                CompOxiDict['N2']           = round(self.getVolPropN2()/self.getVolPropO2(), 3)
                CompProdDict['N2']          = 0.0
                CoeffAbsorpDict['N2']       = self.defaultSpeciesProperties()['coeff_absorption']
                ChemicalFormulaDict['N2']   = 'N2'
            #CO yield
            if self.getCOyield() != 0.0:
                ListOfSpecies.append('CO')
                CompFuelDict['CO']          = 0.0
                CompOxiDict['CO']           = 0.0
                CompProdDict['CO']          = self.getCOyield()
                CoeffAbsorpDict['CO']       = self.defaultSpeciesProperties()['coeff_absorption']
                ChemicalFormulaDict['CO']   = 'CO'
            #CS yield
            if self.getCSyield() != 0.0:
                ListOfSpecies.append('CS')
                CompFuelDict['CS']          = 0.0
                CompOxiDict['CS']           = 0.0
                CompProdDict['CS']          = self.getCSyield()
                CoeffAbsorpDict['CS']       = self.defaultSpeciesProperties()['coeff_absorption']
                ChemicalFormulaDict['CS']   = 'C(S)'
            #H2O
            ListOfSpecies.append('H2O')
            CompFuelDict['H2O']          = 0.0
            CompOxiDict['H2O']           = 0.0
            CompProdDict['H2O']          = 0.0
            CoeffAbsorpDict['H2O']       = self.defaultSpeciesProperties()['coeff_absorption']
            ChemicalFormulaDict['H2O']   = 'H2O'
            #CO2
            if 'C' in str(self.getChemicalFormulaFuel()).upper():
                ListOfSpecies.append('CO2')
                CompFuelDict['CO2']          = 0.0
                CompOxiDict['CO2']           = 0.0
                CompProdDict['CO2']          = 0.0
                CoeffAbsorpDict['CO2']       = self.defaultSpeciesProperties()['coeff_absorption']
                ChemicalFormulaDict['CO2']   = 'CO2'
            #HCl
            if 'CL' in str(self.getChemicalFormulaFuel()).upper():
                ListOfSpecies.append('HCl')
                CompFuelDict['HCl']          = 0.0
                CompOxiDict['HCl']           = 0.0
                CompProdDict['HCl']          = 0.0
                CoeffAbsorpDict['HCl']       = self.defaultSpeciesProperties()['coeff_absorption']
                ChemicalFormulaDict['HCl']   = 'HCl'


        #Open the Thermochemistry Data File
        f = open(file_path, "w")

        #Check if all the species have a given composition with the "user" mode
        self.Error_GUI = False
        if Option_UserMode == 'user':
            CheckComp = {}
            FLAG_No_Oxi = False 
            FLAG_No_Fuel = False 
            FLAG_No_Prod = False 
            FLAG_No_compo = False

            if not ListOfSpecies :
                f.write(ErrorLine['Error'])
                f.write(ErrorLine['No_species'])
                f.close()
                self.Error_GUI = True
                return

            for label in ListOfSpecies:
                CheckComp[label] = CompFuelDict[label]+CompOxiDict[label]+CompProdDict[label]

            if max(CompOxiDict.values()) == 0 : FLAG_No_Oxi = True
            if max(CompFuelDict.values()) == 0 : FLAG_No_Fuel = True
            if max(CompProdDict.values()) == 0 : FLAG_No_Prod = True
            if 0 in CheckComp.values(): FLAG_No_compo = True

            if FLAG_No_Oxi | FLAG_No_Fuel | FLAG_No_Prod | FLAG_No_compo :
                f.write(ErrorLine['Error'])
                if FLAG_No_Oxi : f.write(ErrorLine['FLAG_No_Oxi'])
                if FLAG_No_Fuel : f.write(ErrorLine['FLAG_No_Fuel'])
                if FLAG_No_Prod : f.write(ErrorLine['FLAG_No_Prod'])
                if FLAG_No_compo :
                    for label in ListOfSpecies:
                        if CheckComp.get(label) == 0 :
                            f.write("The " + label + " " + ChemicalFormulaDict[label]+ ErrorLine['FLAG_No_compo'])
                f.close()
                self.Error_GUI = True
                return

        # Write the number of species, and parameters for the tabulation ENTH TEMP
        NumberOfSpecies = len(ListOfSpecies)
        NbPointsTabu = self.getNbPointsTabu()
        MinimumTemp  = self.getMinimumTemp()
        MaximumTemp  = self.getMaximumTemp()
        f.write('{:<10}'.format(str(NumberOfSpecies))+InfoLine['NumberOfSpecies']+"\n")
        f.write('{:<10}'.format(str(NbPointsTabu))+InfoLine['NbPointsTabu']+"\n")
        f.write('{:<10}'.format(str(MinimumTemp))+InfoLine['MinimumTemp']+"\n")
        f.write('{:<10}'.format(str(MaximumTemp))+InfoLine['MaximumTemp']+"\n")
        f.write(InfoLine['LineInfo-GaseousSpecies']+"\n")
        f.write(" "*10)

        # Reorder the species to have the Fuel in first, then Oxidiser and 
        # finish by the Product and sorted the compositions (see colecd.f90)
        maxCompFuel = 0.0
        maxCompOxi  = 0.0
        maxCompProd = 0.0
        for label in ListOfSpecies:
            maxCompFuel = max(maxCompFuel, CompFuelDict[label])
            maxCompOxi = max(maxCompOxi, CompOxiDict[label])
            maxCompProd = max(maxCompProd, CompOxiDict[label])

        Order = []
        for i in range(NumberOfSpecies):
            Order.append(3)

        i = 0
        for label in ListOfSpecies:
            if CompFuelDict[label] > 0.0:
                Order[i] = 0 + CompFuelDict[label]/maxCompFuel
            if CompOxiDict[label] > 0.0:
                Order[i] = 1 + CompOxiDict[label]/maxCompOxi
            if CompProdDict[label] > 0.0:
                Order[i] = 3 + CompProdDict[label]/maxCompProd
            i = i + 1
      
        getSpeciesNamesList_Sorted = ListOfSpecies
        Order, getSpeciesNamesList_Sorted = zip(*sorted(zip(Order, getSpeciesNamesList_Sorted)))

        # Write the chemical formula of all the species
        for label in getSpeciesNamesList_Sorted:
            f.write('{:>15}'.format(str(ChemicalFormulaDict[label])))
        f.write("\n")
        f.write(" "*10)

        # Write the absoption coefficient for all the species
        for label in getSpeciesNamesList_Sorted:
            f.write('{:>15}'.format(str(CoeffAbsorpDict[label])))
        f.write(" "*5+InfoLine['CoeffAbsorp']+"\n")

        # List of all chemical formula
        LstChemicalFormula = []
        for label in getSpeciesNamesList_Sorted:
            LstChemicalFormula.append(str(ChemicalFormulaDict[label]))

        # Write the number of chemical element (CHON)
        GlobalElemCompo = {}
        ChemicalElem_sorted = sorted(self.ChemicalElem, key=len, reverse=True)
        for Elem in ChemicalElem_sorted:
            GlobalElemCompo[Elem] = False
        for Elem in ChemicalElem_sorted:
            if Elem in str(LstChemicalFormula).upper():
                GlobalElemCompo[Elem] = True
        f.write('{:<15}'.format(str(str(GlobalElemCompo).count("True")))+" "*15*NumberOfSpecies+InfoLine['GlobalElemCompo']+"\n")

        # Create a dictionnary with the composition in chemical elements for each species
        Composition = {}
        for label in getSpeciesNamesList_Sorted:
            Composition[label] = self.getNumberOfChemicalElem(ChemicalFormulaDict[label])

        # Write the number of chemical element (CHON) for each species
        for Elem in self.ChemicalElem:
            if GlobalElemCompo[Elem] == True:
                f.write('{:<10}'.format(str(self.MolarMass.get(Elem))))
                for label in getSpeciesNamesList_Sorted:
                    f.write('{:>15}'.format(str(Composition[label][Elem])))
                if Elem == self.ChemicalElem[0]:
                    f.write(" "*5+InfoLine['LineInfo-ChemElem'])
                f.write('\n')

        # Write the number of known global species (always 2 (Fuel and Oxy) for the moment)
        f.write('{:<15}'.format(str(self.NumberOfKnownGlobalSpecies))+" "*15*NumberOfSpecies+InfoLine['NumberOfKnownGlobalSpecies']+"\n")
        f.write(" "*10)

        #Write the Fuel composition
        for label in getSpeciesNamesList_Sorted:
            f.write('{:>15}'.format(str(CompFuelDict[label])))
        f.write(" "*5+InfoLine['FuelComposition']+"\n")
        f.write(" "*10)

        #Write the Oxidiser composition
        for label in getSpeciesNamesList_Sorted:
            f.write('{:>15}'.format(str(CompOxiDict[label])))
        f.write(" "*5+InfoLine['OxiComposition']+"\n")
        f.write(" "*10)

        #Write the Product composition
        for label in getSpeciesNamesList_Sorted:
            f.write('{:>15}'.format(str(CompProdDict[label])))
        f.write(" "*5+InfoLine['ProdComposition'])
        
        #If all the species have a given composition : ngazg == 3
        if self.NumberOfKnownGlobalSpecies == 3 :
            position_fuel = 1
            position_oxi = 2
            f.write("\n")
            f.write('{:<15}'.format(str(self.NbReaction))+" "*15*NumberOfSpecies+InfoLine['NbReaction']+"\n")
            f.write('{:<15}'.format(str(position_fuel)))
            f.write('{:<15}'.format(str(position_oxi)))

            Prodsum = {}
            Prodsum['Fuel'] = 0.0
            Prodsum['Oxi'] = 0.0
            Prodsum['Prod'] = 0.0
            global_stoeg = {}
            global_stoeg['Fuel'] = 0.0
            global_stoeg['Oxi'] = 0.0
            global_stoeg['Prod'] = 0.0

            for label in getSpeciesNamesList_Sorted:
                Prodsum['Fuel'] = Prodsum['Fuel'] + float(Composition[label]['O'])*CompFuelDict[label]
                Prodsum['Oxi'] = Prodsum['Oxi'] + float(Composition[label]['O'])*CompOxiDict[label]
                Prodsum['Prod'] = Prodsum['Prod'] + float(Composition[label]['O'])*CompProdDict[label]
                global_stoeg['Fuel'] = global_stoeg['Fuel'] + CompFuelDict[label]
                global_stoeg['Oxi'] = global_stoeg['Oxi'] + CompOxiDict[label]
                global_stoeg['Prod'] = global_stoeg['Prod'] + CompProdDict[label]

            nreact_oxi = (Prodsum['Prod']-Prodsum['Fuel'])/Prodsum['Oxi']
            global_stoeg['Oxi'] = -global_stoeg['Oxi']*nreact_oxi
            f.write('{:<15}'.format(str(-global_stoeg['Fuel'])))
            f.write('{:<15}'.format(str(global_stoeg['Oxi'])))
            f.write('{:<15}'.format(str(global_stoeg['Prod'])))
            f.write(" "*15*(max(1,NumberOfSpecies-4))+InfoLine['Stoeg']+"\n")

        #Close the Thermochemistry Data File
        f.close()

#-------------------------------------------------------------------------------
# End
#-------------------------------------------------------------------------------
